#!/usr/bin/env python
""" Provides functions to load and write models, media definitions and the manual annotation table

Depending on the application the model needs to be loaded with cobra (memote)
or with libSBML (activation of groups). The media definitions are denoted in a csv within the data folder of this repository, thus the functions will only work if the user clones the repository. The manual_annotations table has to follow the specific layout given in the data folder in order to work with this module.
"""

import cobra
import click
import yaml
import os
import re
import gffutils
import sqlalchemy
import logging
import pandas as pd
from cobra import Model as cobraModel
from ols_client import EBIClient
from Bio import Entrez, SeqIO
from refinegems.databases import PATH_TO_DB, initialise_database
from libsbml import Model as libModel
from libsbml import SBMLReader, writeSBMLToFile, SBMLValidator, SBMLDocument
from datetime import date

__author__ = "Famke Baeuerle and Gwendolyn O. Gusak"


def load_model_cobra(modelpath: str) -> cobraModel:
    """Loads model using COBRApy

    Args:
        - modelpath (str): Path to GEM

    Returns:
        cobraModel: Loaded model by COBRApy
    """
    mod = cobra.io.read_sbml_model(modelpath)
    return mod


def load_model_libsbml(modelpath: str) -> libModel:
    """Loads model using libSBML

    Args:
        - modelpath (str): Path to GEM

    Returns:
        libModel: loaded model by libSBML
    """
    reader = SBMLReader()
    read = reader.readSBMLFromFile(modelpath)  # read from file
    mod = read.getModel()
    return mod


def load_multiple_models(models: list[str], package: str) -> list:
    """Loads multiple models into a list

    Args:
        - models (list): List of paths to models
        - package (str): COBRApy|libSBML

    Returns:
        list: List of model objects loaded with COBRApy|libSBML
    """
    loaded_models = []
    for modelpath in models:
        if package == 'cobra':
            loaded_models.append(load_model_cobra(modelpath))
        elif package == 'libsbml':
            loaded_models.append(load_model_libsbml(modelpath))
    return loaded_models


def load_document_libsbml(modelpath: str) -> SBMLDocument:
    """Loads model document using libSBML

    Args:
        - modelpath (str): Path to GEM

    Returns:
        SBMLDocument: Loaded document by libSBML
    """
    reader = SBMLReader()
    read = reader.readSBMLFromFile(modelpath)  # read from file
    return read


def load_medium_custom(mediumpath: str) -> pd.DataFrame:
    """Helper function to read medium csv

    Args:
        - mediumpath (str): path to csv file with medium

    Returns:
        pd.DataFrame: Table of csv
    """
    medium = pd.read_csv(mediumpath, sep=';')
    medium['BiGG_R'] = 'R_EX_' + medium['BiGG'] + '_e'
    medium['BiGG_EX'] = 'EX_' + medium['BiGG'] + '_e'
    return medium


def load_medium_from_db(mediumname: str) -> pd.DataFrame:
    """Wrapper function to extract subtable for the requested medium from the database 'data.db'

    Args:
        - mediumname (str): Name of medium to test growth on

    Returns:
        pd.DataFrame: Table containing composition for one medium with metabs added as BiGG_EX exchange reactions
    """
    medium_query = f"SELECT * FROM media m JOIN media_compositions mc ON m.id = mc.medium_id WHERE m.medium = '{mediumname}'" 
    medium = load_a_table_from_database(medium_query)
    medium = medium[['medium', 'medium_description', 'BiGG', 'substance']]
    medium['BiGG_R'] = 'R_EX_' + medium['BiGG'] + '_e'
    medium['BiGG_EX'] = 'EX_' + medium['BiGG'] + '_e'
    return medium


def load_all_media_from_db(mediumpath: str) -> pd.DataFrame: 
    """Helper function to extract media definitions from media_db.csv

    Args:
        - mediumpath (str): Path to csv file with medium database

    Returns:
        pd.DataFrame: Table from csv with metabs added as BiGG_EX exchange reactions
    """
    media = pd.read_csv(mediumpath, sep=';')
    media['BiGG_R'] = 'R_EX_' + media['BiGG'] + '_e'
    media['BiGG_EX'] = 'EX_' + media['BiGG'] + '_e'

    media['group'] = media['medium'].ne(media['medium'].shift()).cumsum()
    grouped = media.groupby('group')
    media_dfs = []
    for name, data in grouped:
        media_dfs.append(data.reset_index(drop=True))
    return media_dfs


def load_manual_annotations(tablepath: str='data/manual_curation.xlsx', sheet_name: str='metab') -> pd.DataFrame:
    """Loads metabolite sheet from manual curation table

    Args:
        - tablepath (str): Path to manual curation table. Defaults to 'data/manual_curation.xlsx'.
        - sheet_name (str): Sheet name for metabolite annotations. Defaults to 'metab'.

    Returns:
        pd.DataFrame: Table containing specified sheet from Excel file
    """
    man_ann = pd.read_excel(tablepath, sheet_name)
    return man_ann


def load_a_table_from_database(table_name_or_query: str) -> pd.DataFrame:
    """| Loads the table for which the name is provided or a table containing all rows for which the query evaluates to 
       | true from the refineGEMs database ('data/database/data.db')

    Args:
        - table_name_or_query (str): Name of a table contained in the database 'data.db'/ a SQL query

    Returns:
        pd.DataFrame: Containing the table for which the name was provided from the database 'data.db'
    """
    sqlalchemy_engine_input = f'sqlite:///{PATH_TO_DB}'
    engine = sqlalchemy.create_engine(sqlalchemy_engine_input)
    open_con = engine.connect()
    
    db_table = pd.read_sql(table_name_or_query, open_con)
    
    open_con.close()
    return db_table


def load_manual_gapfill(tablepath: str='data/manual_curation.xlsx' , sheet_name: str='gapfill') -> pd.DataFrame:
    """Loads gapfill sheet from manual curation table

    Args:
        - tablepath (str): Path to manual curation table. Defaults to 'data/manual_curation.xlsx'.
        - sheet_name (str): Sheet name for reaction gapfilling. Defaults to 'gapfill'.

    Returns:
        pd.DataFrame: Table containing sheet with name 'gapfill'|specified sheet_name from Excel file
    """
    man_gapf = pd.read_excel(tablepath, sheet_name)
    return man_gapf


def write_to_file(model: libModel, new_filename: str):
    """Writes modified model to new file

    Args:
        - model (libModel): Model loaded with libSBML
        - new_filename (str): Filename|Path for modified model
    """
    new_document = model.getSBMLDocument()
    writeSBMLToFile(new_document, new_filename)
    logging.info("Modified model written to " + new_filename)


def write_report(dataframe: pd.DataFrame, filepath: str):
    """Writes reports stored in dataframes to xlsx file

    Args:
        - dataframe (pd.DataFrame): Table containing output
        - filepath (str): Path to file with filename
    """
    writer = pd.ExcelWriter(str(os.path.abspath('.')) + '/' + filepath)
    dataframe.to_excel(writer)
    writer.save()


def validate_libsbml_model(model: libModel) -> int:
    """Debug method: Validates a libSBML model with the libSBML validator
    Args:
        - model (libModel): A libSBML model
        
    Returns:
        int: Integer specifying if vaidate was successful or not
    """
    validator = SBMLValidator()
    doc = model.getSBMLDocument()
    
    return validator.validate(doc)


def parse_fasta_headers(filepath: str, id_for_model: bool=False) -> pd.DataFrame:
    """Parses FASTA file headers to obtain:
    
        - the protein_id
        - and the model_id (like it is obtained from CarveMe)
            
    corresponding to the locus_tag
        
    Args:
        - filepath (str): Path to FASTA file
        - id_for_model (bool): True if model_id similar to autogenerated GeneProduct ID should be contained in resulting table
        
    Returns:
        pd.DataFrame: Table containing the columns locus_tag, Protein_id & Model_id
    """
    keyword_list = ['protein', 'locus_tag']
    tmp_dict = dict()
    if id_for_model:
        locus2ids = {
            'locus_tag': [],
            'protein_id': [],
            'model_id': [],
            'name': []
        }
    else:
        locus2ids = {
            'locus_tag': [],
            'protein_id': [],
            'name': []
        }
   
    with open(filepath, 'r') as handle:
        for record in SeqIO.parse(handle, 'fasta'):
            header = record.description
            protein_id = record.id.split('|')[1].split('prot_')[1].split('.')[0].strip()
            descriptors = re.findall('\[+(.*?)\]', header)
            if id_for_model:
                model_id = re.sub("\||\.", "_", record.id)
                model_id = f'G_{model_id}'
         
            descriptors.insert(0, protein_id)
            
            tmp_dict['protein_id'] = str(protein_id)
            
            for entry in descriptors:
                entry = entry.strip().split('=')
               
                if entry[0] in keyword_list:
                    if entry[0] == 'protein_id':
                        tmp_dict[entry[0]] = entry[1].split('.')[0]
                    else:
                        tmp_dict[entry[0]] = entry[1]
                    
            locus2ids.get('locus_tag').append(tmp_dict.get('locus_tag'))
            locus2ids.get('protein_id').append(tmp_dict.get('protein_id'))
            locus2ids.get('name').append(tmp_dict.get('protein'))
            if id_for_model:
                locus2ids.get('model_id').append(model_id)
            
    return pd.DataFrame(locus2ids)


def search_ncbi_for_gpr(locus: str) -> str:
    """Fetches protein name from NCBI

    Args:
        - locus (str): NCBI compatible locus_tag

    Returns:
        str: Protein name|description
    """
    handle = Entrez.efetch(
        db="protein",
        id=locus,
        rettype="gbwithparts",
        retmode='text')
    records = SeqIO.parse(handle, "gb")

    for i, record in enumerate(records):
        if (locus[0] == 'W'):
            return record.description, locus
        else:
            for feature in record.features:
                if feature.type == "CDS":
                    return record.description, feature.qualifiers["locus_tag"][0]


def parse_gff_for_gp_info(gff_file: str) -> pd.DataFrame:
    """Parses gff file of organism to find gene protein reactions based on locus tags

    Args:
        - gff_file (str): Path to gff file of organism of interest

    Returns:
        pd.DataFrame: Table containing mapping from locus tag to GPR
    """
    db = gffutils.create_db(
        gff_file,
        ':memory:',
        merge_strategy='create_unique')
    mapping_cds = {}
    for feature in db.all_features():
        attr = dict(feature.attributes)
        try:
            if str(attr['gbkey'][0]) == 'CDS':
                mapping_cds[attr['Name'][0]] = attr['Parent'][0]
        except BaseException:
            pass
    mapping_df = pd.DataFrame.from_dict(
        mapping_cds,
        columns=['Parent'],
        orient='index').reset_index().rename(
        columns={
            'index': 'GPR'})

    def extract_locus(feature):
        try:
            return db[feature].attributes['old_locus_tag'][0]
        except BaseException:
            pass
        return None

    mapping_df['locus_tag'] = mapping_df.apply(
        lambda row: extract_locus(row['Parent']), axis=1)
    return mapping_df.drop('Parent', axis=1)


def search_sbo_label(sbo_number: str) -> str:
    """Looks up the SBO label corresponding to a given SBO Term number

    Args:
        - sbo_number (str): Last three digits of SBO-Term as str

    Returns:
        str: Denoted label for given SBO Term
    """
    sbo_number = str(sbo_number)
    client = EBIClient()
    sbo = client.get_term('sbo', 'http://biomodels.net/SBO/SBO_0000' + sbo_number)
    return sbo['_embedded']['terms'][0]['label']


def save_user_input(configpath: str) -> dict[str: str]:
    """This aims to collect user input from the command line to create a config file, 
    will also save the user input to a config if no config was given

    Args:
        - configpath (str): Path to config file if present
        
    Returns:
        dict: Either loaded config file or created from user input
    """
    if os.path.isfile(configpath):
        with open(configpath) as f:
            config = yaml.safe_load(f)
        print(config)
        return config
    else:
        print('No config or no valid config given, you will be asked for input')
        user_input = {}
        
        update_db = click.confirm('Do you want to update the database?')
        user_input['db_update'] = update_db
        
        if update_db:
            initialise_database()

        out_path = click.confirm('Do you want to keep the output path "../rg_out/"?', default=True)
        if not out_path:
            user_input['out_path'] = click.prompt('Enter your desired output path')
        else:
            user_input['out_path'] = '../rg_out/'
        
        user_input['visualize'] = click.confirm('Do you want to generate visualizations of your model(s)?')
            
        growth_basis = click.prompt('Enter the base uptakes for growth simulation (d for default_uptake, m for minimal_uptake)')
        if growth_basis == 'd':
            user_input['growth_basis'] = 'default_uptake'
        if growth_basis == 'm':
            user_input['growth_basis'] = 'minimal_uptake'
            
        user_input['anaerobic_growth'] = click.confirm('Do you want to simulate anaerobic growth?')
        
        multiple = click.confirm('Do you want to simulate and compare multiple models?')
        user_input['multiple'] = multiple
        if multiple:
            list_of_models = []
            while True:
                file_path = click.prompt('Enter file path to model (or "stop" to stop)')
                if file_path.lower() == 'stop':
                    break
                elif os.path.isfile(file_path):
                    list_of_models.append(file_path)
                    print('Added file:', file_path)
                else:
                    print('File does not exist. Please enter a valid file path.')
            print('The following models will be compared:')
            print(list_of_models)
            user_input['multiple_paths'] = list_of_models
        possible_media = load_a_table_from_database('media')['medium'].to_list()
        possible_media_str = '|'.join(possible_media)
        list_of_media = []
        while True:
            medium = click.prompt(f'Enter medium to simulate growth on ({possible_media_str}) (or "stop" to stop)')
            if medium.lower() == 'stop':
                break
            elif medium in possible_media:
                if medium not in list_of_media:
                    list_of_media.append(medium)
                else:
                    print(medium + ' is already in the list.')
            else:
                print('Please choose a medium from the given list.')
        user_input['media'] = list_of_media
        
        single = click.confirm('Do you want to investigate or curate a single model?')
        user_input['single'] = single
        if single:
            not_valid = True
            while not_valid:
                model = click.prompt('Path to your model file.')
                if os.path.isfile(model):
                    user_input['model'] = model
                    not_valid = False
                else:
                    print('File does not exist. Please enter a valid file path')
        
            user_input['memote'] = click.confirm('Do you want to run MEMOTE (takes some time)?')    
            user_input['modelseed'] = click.confirm('Do you want to compare your model entities to the ModelSEED database?')
        
            gap_analysis = click.confirm('Do you want to run a gap analysis?') 
            user_input['gap_analysis'] = gap_analysis
            if gap_analysis:
                gap_analysis_params = {}
                db_to_compare = click.prompt('One of the choices KEGG|BioCyc|KEGG+BioCyc') #|GFF
                gap_analysis_params['db_to_compare'] = db_to_compare
                if db_to_compare == 'KEGG' or db_to_compare == 'KEGG+BioCyc':
                    gap_analysis_params['organismid'] = click.prompt('Enter the KEGG Organism ID')
                    gap_analysis_params['gff_file'] = click.prompt('Enter the path to your organisms RefSeq GFF file')
                if db_to_compare == 'BioCyc' or db_to_compare == 'KEGG+BioCyc':
                    Path0 = click.prompt('Enter the path to your BioCyc TXT file containing a SmartTable with the columns \'Accession-2\' and \'Reaction of gene\'')
                    Path1 = click.prompt('Enter the path to your BioCyc TXT file containing a SmartTable with all reaction relevant information')
                    Path2 = click.prompt('Enter the path to your Biocyc TXT file containing a SmartTable with all metabolite relevant information')
                    Path3 = click.prompt('Enter path to protein FASTA file used as input for CarveMe')
                    gap_analysis_params['biocyc_files'] = [Path0, Path1, Path2, Path3]
                user_input['gap_analysis_params'] = gap_analysis_params
                
            mod = click.confirm('Do you want to use functions to modify your model?')
            if mod:
                
                new_path = click.confirm('Do you want to save your modified model to ' + user_input['out_path'] + '<model.id>_modified_<today>.xml?')
                if new_path:
                    user_input['model_out'] = 'stdout'
                else:
                    user_input['model_out'] = click.prompt('Enter path and filename to where to save the modified model')
                
                user_input['gapfill_model'] = click.confirm('Do you want to fill gaps in your model?')
                
                if not gap_analysis:
                    user_input['gap_analysis_file'] = click.prompt('Enter path to Excel file with which gaps should be filled')
                
                user_input['keggpathways'] = click.confirm('Do you want to add KEGG Pathways?')
                    
                user_input['sboterms'] = click.confirm('Do you want to update the SBO Terms?')
                
                user_input['charge_corr'] = click.confirm('Do you want to add charges to uncharged metabolites?')
                    
                man_cur = click.confirm('Do you want to modify your model with the manual curations table?')
                user_input['man_cur'] = man_cur

                if man_cur:
                    entrez_email = click.prompt('Email to access NCBI Entrez')
                    user_input['entrez_email'] = entrez_email
                    man_cur_type = click.prompt('Enter type of curation (gapfill|metabs)')
                    user_input['man_cur_type'] = man_cur_type
                    man_cur_table = click.prompt('Enter the path to the manual curations table')
                    user_input['man_cur_table'] = man_cur_table

                polish = click.confirm('Do you want to polish the model?')
                user_input['polish'] = polish

                if polish:
                    entrez_email = click.prompt('Email to access NCBI Entrez')
                    user_input['entrez_email'] = entrez_email
                    id_db = click.prompt('What database is your model based on? BIGG|VMH')
                    user_input['id_db'] = id_db
                    lab_strain = not click.confirm('Does your modeled organism have a database entry?', default=True)
                    user_input['lab_strain'] = lab_strain
                    protein_fasta = click.prompt('If possible, provide the path to your Protein FASTA file used for CarveMe')
                    user_input['protein_fasta'] = protein_fasta
                    
            else:
                user_input['keggpathways'] = False
                user_input['polish'] = False
                user_input['sboterms'] = False
                user_input['charge_corr'] = False
                user_input['gapfill_model'] = False
                user_input['man_cur'] = False
            
        today = date.today().strftime("%Y%m%d")
        
        print('This is your input:')
        print(user_input)
        if not os.path.isdir(user_input['out_path']):
            print('Given out_path is not yet a directory, creating ' + user_input['out_path'])
            os.makedirs(user_input['out_path'])
        with open(user_input['out_path'] + 'user_input_' + str(today) + '.yaml', 'w') as f:
            yaml.dump(user_input, f)
        print('Your input was saved as yaml to '+ user_input['out_path'] + 'user_input_' + str(today) + '.yaml')
        return user_input


# declares public API for this module
# loads functions from subscripts which are needed in main.py
import refinegems.databases
import refinegems.growth
import refinegems.modelseed
import refinegems.sboann
import refinegems.charges
import refinegems.pathways
import refinegems.curate
import refinegems.polish
import refinegems.cvterms
import refinegems.analysis_db
import refinegems.analysis_biocyc
import refinegems.analysis_kegg
import refinegems.entities
import refinegems.gapfill
import refinegems.comparison
import refinegems.investigate
import refinegems.io

__author__ = "Famke Baeuerle"



#!/usr/bin/env python
import io
import re
import sqlite3
import requests
import pandas as pd
from enum import Enum
from sqlite3 import Error
from os import path

__author__ = 'Gwendolyn O. Gusak'


PATH_TO_DB_DATA = path.join(path.abspath(path.dirname(path.dirname(__file__))), 'data/database')
PATH_TO_DB = path.join(PATH_TO_DB_DATA, 'data.db')
VERSION_FILE = path.join(PATH_TO_DB_DATA, 'current_bigg_db_version.txt')
VERSION_URL = 'http://bigg.ucsd.edu/api/v2/database_version'

class ValidationCodes(Enum):
   """Validation codes for the database

      Args:
         - Enum (Enum): Provided as input to get a number mapping for the codes
   """
   COMPLETE = 0,  # All tables are in data.db
   EMPTY = 1,  # data.db is either empty or incorrect
   BIGG = 2,  # Only BiGG tables are in data.db
   SBO_MEDIA = 3,  # Only SBO & Media tables are in data.db (Can only occurr together) 
   BIGG_SBO_MEDIA = 4,  # Only BiGG, SBO and media tables are in data.db 
   MODELSEED_COMPOUNDS = 5,  # Only ModelSEED compounds table is in data.db
   BIGG_MSEED_COMPPOUNDS = 6,  # Only Bigg and ModelSEED compounds tables are in data.db
   SBO_MEDIA_MSEED_COMPOUNDS = 7  # Only SBO, media and ModelSEED compounds tables are in data.db


def is_valid_database(db_cursor: sqlite3.Cursor) -> int:
   """Verifies if database has:
         - 2 tables with names 'bigg_metabolites' & 'bigg_reactions'
         - 2 tables with names 'bigg_to_sbo' & 'ec_to_sbo'
         - 2 tables with names 'media' & 'media_composition'
         - 1 table with name 'modelseed_compounds'
   
   Args:
      - db_cursor (sqlite3.Cursor): Cursor from open connection to the database (data.db)

   Returns:
      int: Corresponding to one of the ValidationCodes
   """
   print('Verifying database...')
   
   # Fetches the table names as string tuples from the connected database
   db_cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
   tables = [string[0] for string in db_cursor.fetchall()]
   
   bigg_tables_contained = len([s for s in tables if re.match('^bigg_(?!to)(.*?)', s, re.IGNORECASE)]) == 2
   sbo_tables_contained = len([s for s in tables if re.match('(.*?)_sbo$', s, re.IGNORECASE)]) == 2
   media_tables_contained = len([s for s in tables if re.match('media', s, re.IGNORECASE)]) == 2
   sbo_media_tables_contained = sbo_tables_contained and media_tables_contained  # These can only occur together
   modelseed_cmpd_tbl_contained = len([s for s in tables if s == 'modelseed_compounds']) == 1
   
   bigg_sbo_media_tbls_contained = bigg_tables_contained and sbo_media_tables_contained
   bigg_modelseed_cmpd_tbls_contained = bigg_tables_contained and modelseed_cmpd_tbl_contained
   sbo_media_modelseed_cmpd_tbls_contained = sbo_media_tables_contained and modelseed_cmpd_tbl_contained
   all_tables_contained = bigg_sbo_media_tbls_contained and modelseed_cmpd_tbl_contained
   
   if all_tables_contained: return ValidationCodes.COMPLETE
   elif bigg_modelseed_cmpd_tbls_contained: return ValidationCodes.BIGG_MSEED_COMPOUNDS
   elif sbo_media_modelseed_cmpd_tbls_contained: return ValidationCodes.SBO_MEDIA_MSEED_COMPOUNDS
   elif bigg_sbo_media_tbls_contained: return ValidationCodes.BIGG_SBO_MEDIA
   elif bigg_tables_contained: return ValidationCodes.BIGG
   elif sbo_media_tables_contained: return ValidationCodes.SBO_MEDIA
   elif modelseed_cmpd_tbl_contained: return ValidationCodes.MODELSEED_COMPOUNDS
   else: return ValidationCodes.EMPTY


def create_sbo_media_database(db_cursor: sqlite3.Cursor):
   """Creates the SBO annotation database with 2 tables ('bigg_to_sbo' & 'ec_to_sbo')
      & the media database with 2 tables ('media', 'media_compositions') from file './data/database/sbo_media_db.sql'

   Args:
      - db_cursor (sqlite3.Cursor): Cursor from open connection to the database (data.db)
   """
   print('Adding SBO and media tables...')
   
   with open(path.join(PATH_TO_DB_DATA, 'sbo_media_db.sql')) as schema:
      db_cursor.executescript(schema.read())


def update_bigg_db(latest_version: str, db_connection: sqlite3.Connection):
   """Updates the BiGG tables 'bigg_metabolites' & 'bigg_reactions' within a database (data.db)

   Args:
      - latest_version (str): String containing the latest version of the BiGG database
      - db_connection (sqlite3.Connection): Open connection to the database (data.db)
   """
   print('Adding BiGG tables...')
   db_connection.execute('DROP TABLE IF EXISTS bigg_metabolites')
   db_connection.execute('DROP TABLE IF EXISTS bigg_reactions')
   
   # Store currently used version
   with open(VERSION_FILE, 'w') as file:
      file.write(latest_version)
   
   # Create BiGG metabolites table
   BIGG_MODELS_METABS_URL = 'http://bigg.ucsd.edu/static/namespace/bigg_models_metabolites.txt'
   bigg_models_metabs = requests.get(BIGG_MODELS_METABS_URL).text
   bigg_models_metabs_df = pd.read_csv(io.StringIO(bigg_models_metabs), dtype=str, sep='\t')
   bigg_models_metabs_df.to_sql('bigg_metabolites', db_connection, index=False)

   # Create BiGG reactions table
   BIGG_MODELS_REACS_URL = 'http://bigg.ucsd.edu/static/namespace/bigg_models_reactions.txt'
   bigg_models_reacs = requests.get(BIGG_MODELS_REACS_URL).text
   bigg_models_reacs_df = pd.read_csv(io.StringIO(bigg_models_reacs), dtype=str, sep='\t')
   bigg_models_reacs_df.to_sql('bigg_reactions', db_connection, index=False)
   

def get_latest_bigg_databases(db_connection: sqlite3.Connection, is_missing: bool=True):
   """Gets the latest BiGG tables for metabolites & reactions if:
         - No version file is locally available
         - The version in the local version file is NOT the latest
         - No BiGG tables currently exist in the database

   Args:
      - db_connection (sqlite3.Connection): Open connection to the database (data.db)
      - is_missing (bool, optional): True if no BiGG tables are in the database. Defaults to True.
   """
   # Check if BiGG database had an update
   LATEST_VERSION = requests.get(VERSION_URL).json()['bigg_models_version']
   
   if not path.exists(VERSION_FILE) or is_missing:
      update_bigg_db(LATEST_VERSION, db_connection)
      
   else:
      with open(VERSION_FILE, 'r') as file:
         version = file.readline().strip()
         
      if version != LATEST_VERSION:
         update_bigg_db(LATEST_VERSION, db_connection)
 
 
def get_modelseed_compounds_database(db_connection: sqlite3.Connection):
   """Retrieves the compounds table from ModelSEED from the respective GitHub repository

   Args:
      - db_connection (sqlite3.Connection): Open connection to the database (data.db)
   """
   print('Adding the ModelSEED compounds table...')
   MODELSEED_COMPOUNDS_URL = 'https://raw.githubusercontent.com/ModelSEED/ModelSEEDDatabase/master/Biochemistry/compounds.tsv'
   modelseed_compounds = requests.get(MODELSEED_COMPOUNDS_URL).text
   modelseed_df = pd.read_csv(io.StringIO(modelseed_compounds), sep='\t')
   modelseed_df.to_sql('modelseed_compounds', db_connection, index=False, if_exists='replace')
    
         
def initialise_database():
   """Initialises/updates the database (data.db)

      After initialisation the database contains:
         - 2 tables with names 'bigg_metabolites' & 'bigg_reactions'
         - 2 tables with names 'bigg_to_sbo' & 'ec_to_sbo'
         - 2 tables with names 'media' & 'media_composition'
         - 1 table with name 'modelseed_compounds' 
   """
   # Initialise empty connection
   con = None

   print('Initialising database...')
   
   # Try to open connection & get cursor
   try:
      con = sqlite3.connect(PATH_TO_DB)
      cursor = con.cursor()
      
      validity_code = is_valid_database(cursor)
      
      if validity_code == ValidationCodes.BIGG:
         print('Only BiGG tables contained in database.')
         create_sbo_media_database(cursor)
         get_modelseed_compounds_database(con)
      
      elif validity_code == ValidationCodes.SBO_MEDIA:
         print('Only SBO and media tables contained in database.')
         get_latest_bigg_databases(con)
         get_modelseed_compounds_database(con)
         
      elif validity_code == ValidationCodes.MODELSEED_COMPOUNDS:
         print('Only ModelSEED compounds table contained in database.')
         create_sbo_media_database(cursor)
         get_latest_bigg_databases(con)
         
      elif validity_code == ValidationCodes.BIGG_SBO_MEDIA:
         print('Only BiGG, SBO and media tables contained in database.')
         get_modelseed_compounds_database(con)
         
      elif validity_code == ValidationCodes.BIGG_MSEED_COMPPOUNDS:
         print('Only BiGG and ModelSEED compounds tables contained in database.')
         create_sbo_media_database(cursor)
         
      elif validity_code == ValidationCodes.SBO_MEDIA_MSEED_COMPOUNDS:
         print('Only SBO, media and ModelSEED compounds tables contained in database.')
         get_latest_bigg_databases(con)
         
      elif validity_code == ValidationCodes.EMPTY:
         print('Incorrect or empty database. Initialise database with required tables...')
         create_sbo_media_database(cursor)
         get_latest_bigg_databases(con)
         get_modelseed_compounds_database(con)
         
      elif validity_code == ValidationCodes.COMPLETE:
         print('Verifying if BiGG tables are up-to-date...')
         get_latest_bigg_databases(con, False)
      
   except Error as e:
      print(e)
   finally:
      if con:
         print('All tables in database up-to-date. Initialisation complete.')
         con.close()
      


#!/usr/bin/env python
import re
import requests
import sqlite3
import pandas as pd
import numpy as np
from refinegems.io import load_a_table_from_database
from refinegems.databases import PATH_TO_DB
from typing import Literal
from tqdm import tqdm
from ratelimit import limits, sleep_and_retry


__author__ = "Famke Baeuerle and Gwendolyn O. Gusak"


ALL_BIGG_COMPARTMENTS_ONE_LETTER = ('c', 'e', 'p', 'm', 'x', 'r', 'v', 'n', 'g', 'u', 'l', 'h', 'f', 's', 'i', 'w', 'y')
ALL_BIGG_COMPARTMENTS_TWO_LETTER = ('im', 'cx', 'um', 'cm', 'mm')
BIGG_REACTIONS_URL = 'http://bigg.ucsd.edu/api/v2/universal/reactions/'
BIGG_METABOLITES_URL = 'http://bigg.ucsd.edu/api/v2/universal/metabolites/'

COMPARTMENTS = ('c', 'e', 'p')


def get_search_regex(other_db: Literal['KEGG', 'BioCyc'], metabolites: bool) -> str:
    """Retrieves the search regex for BioCyc/KEGG to be used in the BiGG mapping

    Args:
        - other_db (Literal): Specifies if the search regex should be for BioCyc or KEGG
        - metabolites (bool): Is required if one wants to search for KEGG Compound IDs in the bigg_models_metabolites.txt
            
    Returns:
        str: Search regex
    """
    if other_db == 'BioCyc':
        return 'BioCyc: http://identifiers.org/biocyc/META:(.*?);'
    elif other_db == 'KEGG':
        if metabolites:
            return 'KEGG Compound: http://identifiers.org/kegg.compound/(.*?);'
        else:
            return 'KEGG Reaction: http://identifiers.org/kegg.reaction/(.*?);'
        
        
def compare_ids(id1: str, id2: str) -> bool:
    """Compares two strings/IDs & Returns True if one string matches most of the other

    Args:
        - id1 (str): ID 1
        - id2 (str): ID 2

    Returns:
        bool: Indicates if most of one string contained in the other
    """
    id1_split, id2_split, id1_single_comp, id2_single_comp, id1_comp, id2_comp = None, None, None, None, None, None
    
    if '_' in id1: id1_split = re.split('_([a-zA-Z]|[0-9])$', id1)[0]
    if '_' in id2: id2_split = re.split('_([a-zA-Z]|[0-9])$', id2)[0]
    if id1.endswith(ALL_BIGG_COMPARTMENTS_ONE_LETTER): id1_single_comp = id1[:-1]
    if id2.endswith(ALL_BIGG_COMPARTMENTS_ONE_LETTER): id2_single_comp = id2[:-1]
    if id1.endswith(ALL_BIGG_COMPARTMENTS_TWO_LETTER): id1_comp = id1[:-2]
    if id2.endswith(ALL_BIGG_COMPARTMENTS_TWO_LETTER): id2_comp = id2[:-2]
    
    similar_ids = False
    if id1 == id2: similar_ids = True  # Both IDs are same
    
    elif id1_split and id2_split and (id1_split == id2_split): similar_ids = True # Both IDs are same but from different compartments
    elif id2_split and (id1 == id2_split): similar_ids = True # - "" -
    elif id1_split and (id1_split == id2): similar_ids = True # - "" -
    
    elif id1_single_comp and id2_single_comp and (id1_single_comp == id2_single_comp): similar_ids = True
    elif id2_single_comp and (id1 == id2_single_comp): similar_ids = True
    elif id1_single_comp and (id1_single_comp == id2_single_comp): similar_ids = True 
    
    elif id1_comp and id2_comp and (id1_comp == id2_comp): similar_ids = True
    elif id2_comp and (id1 == id2_comp): similar_ids = True
    elif id1_comp and (id1_comp == id2): similar_ids = True
    
    elif id1_split and id2_single_comp and (id1_split == id2_single_comp): similar_ids = True
    elif id2_split and id1_single_comp and (id1_single_comp == id2_split): similar_ids = True
    
    elif id1_split and id2_comp and (id1_split == id2_comp): similar_ids = True
    elif id2_split and id1_comp and (id1_comp == id2_split): similar_ids = True
    
    elif id1_comp and id2_single_comp and (id1_comp == id2_single_comp): similar_ids = True
    elif id2_comp and id1_single_comp and (id1_single_comp == id2_comp): similar_ids = True
    
    else: similar_ids = False

    return similar_ids


def keep_only_reactions_in_certain_compartments(complete_df: pd.DataFrame) -> pd.DataFrame:
    """Extracts all possible BiGG ID variations from database for a BiGG reaction ID, gets the metabolite compartments
        & returns table containing only reactions which happen in one of the provided compartments
        
    Args:
        - complete_df (pd.DataFrame): Table containing at least the columns 'bigg_id' & 'KEGG'/'BioCyc'
        
    Returns:
        pd.DataFrame: Table containing reactions & their compartments
    """
    tqdm.pandas()
    db = 'KEGG' if 'KEGG' in complete_df.columns else 'BioCyc'
    complete_df = complete_df[['bigg_id', db]]  # Remove all unnecessary columns
    
    # (1) Find all occurrencs of a BiGG reaction ID in bigg_reactions table in database
    def get_all_similar_bigg_ids(bigg_id_in: str) -> list[str]:
        
        if '_' in bigg_id_in: bigg_id = re.split('_([a-zA-Z]|[0-9])$', bigg_id_in)[0]
        elif bigg_id_in.endswith(ALL_BIGG_COMPARTMENTS_ONE_LETTER): bigg_id = bigg_id_in[:-1]
        elif bigg_id_in.endswith(ALL_BIGG_COMPARTMENTS_TWO_LETTER): bigg_id = bigg_id_in[:-2]
        else: bigg_id = bigg_id_in
        
        query = f"SELECT bigg_id, INSTR(bigg_id, '{bigg_id}') bi FROM bigg_reactions WHERE bi > 0"
        result = con.execute(query).fetchall()
        result = [result_tuple[0] for result_tuple in result] if result else [bigg_id_in]
        result = [res for res in result if compare_ids(bigg_id, res)]
        return result
    
    # (2) Use list of all BiGG IDs obtained from database table bigg_reactions to get 'metabolites'
    @sleep_and_retry
    @limits(calls=10, period=1)
    def get_reaction_compartment(bigg_id: str) -> str:
        
        metabs_from_reac = requests.get(BIGG_REACTIONS_URL + bigg_id, allow_redirects=False).json()['metabolites']
                
        comps = [comp_dict.get('compartment_bigg_id') for comp_dict in metabs_from_reac]  # Get all compartments for reaction
        contained_in_compartments = [(comp in COMPARTMENTS) for comp in comps]  # Get True for correct compartment        
        if not all(contained_in_compartments):  # At least one compartment not correct
            return np.nan
        else:  # All compartments correct
            if len(set(comps)) == 1:  # Set of found compartments of reaction = 1: Reaction happens in one compartment
                return comps[0]
            else:  # Not so important but do not remove reaction as reaction in correct compartments
                return 'exchange'  # Probably exchange reaction
    
    # Connect to database & get similar IDs (1)
    print('Getting all similar IDs...')
    con = sqlite3.connect(PATH_TO_DB)  # Open connection to database
    complete_df.loc[:,'bigg_id_list'] = complete_df.loc[:, 'bigg_id'].progress_map(get_all_similar_bigg_ids)
    #complete_df.progress_apply(get_all_similar_bigg_ids, axis=1)
    con.close()  # Close connection to database
    
    # Adjust table to contain one BiGG ID per row from bigg_id_list (1)
    complete_df.loc[:, 'id_group'] = complete_df['bigg_id'].ne(complete_df['bigg_id'].shift()).cumsum()  # Group similar IDs
    complete_df.drop(labels='bigg_id', axis=1, inplace=True)  # Drop 'bigg_id' as no longer required
    complete_df = complete_df.explode('bigg_id_list', ignore_index=True)  # Expand 'bigg_id_list' column
    complete_df.rename(columns={'bigg_id_list': 'bigg_id'}, inplace=True)  # Rename 'bigg_id_list' to 'bigg_id'
    
    # (2) Get all compartments for each reaction from BiGG database API
    print(f'Getting all IDs with correct compartment {COMPARTMENTS}...')
    complete_df.loc[:, 'compartment'] = complete_df.loc[:, 'bigg_id'].progress_map(get_reaction_compartment)
    #complete_df.progress_apply(get_reaction_compartment, axis=1)  # (2)
    
    # (3) Remove reactions with compartment = NaN
    complete_df.dropna(subset=['compartment'], inplace=True)
        
    return complete_df

 
# Function originally from refineGEMs.genecomp/refineGEMs.KEGG_analysis --- Modified
def get_bigg2other_db(other_db: Literal['KEGG', 'BioCyc'], metabolites: bool=False) -> pd.DataFrame:
    """Uses list of BiGG reactions/metabolites to get a mapping from BiGG to KEGG/BioCyc Id

    Args:
        - other_db (Literal): Set to 'KEGG'/'BioCyc' to map KEGG/BioCyc IDs to BiGG IDs
        - metabolites (bool): Set to True to map other_db IDs to BiGG IDs for metabolites

    Returns:
        pd.DataFrame: Table containing BiGG Ids with corresponding KEGG/BioCyc Ids
    """
    
    # Get only rows with BioCyc/KEGG entries
    db_table_name = 'bigg_metabolites' if metabolites else 'bigg_reactions'
    reaction_or_compound = 'Compound' if metabolites else 'Reaction'
    other_db_query = other_db if other_db == 'BioCyc' else ' '.join([other_db, reaction_or_compound])
    bigg_db_query = f"SELECT *, INSTR(database_links, '{other_db_query}:') o_db FROM {db_table_name} WHERE o_db > 0"
    bigg_db_df = load_a_table_from_database(bigg_db_query)
    
    db_search_regex = get_search_regex(other_db, metabolites)
    
    def find_other_db(database_links: str):
        m = re.findall(
            db_search_regex,
            str(database_links))
        if m:
            return m
        else:
            return None
    
    bigg_db_df[other_db] = bigg_db_df.apply(
        lambda row: find_other_db(row['database_links']), axis=1)
    bigg_db_df = bigg_db_df.explode(other_db, ignore_index=True)
    
    if not metabolites:
        bigg_db_df = keep_only_reactions_in_certain_compartments(bigg_db_df)
        
    bigg_df = bigg_db_df[['bigg_id', other_db]] if metabolites else bigg_db_df[['bigg_id', other_db, 'compartment', 'id_group']]

    return bigg_df
 
 
# Function originally from refineGEMs.genecomp/refineGEMs.KEGG_analysis --- Modified
def compare_bigg_model(complete_df: pd.DataFrame, model_entities: pd.DataFrame, metabolites: bool=False) -> pd.DataFrame:
    """Compares missing entities obtained through genes extracted via KEGG/BioCyc to entities in the model
        Needed to back check previous comparisons.

    Args:
        - complete_df (pd.DataFrame): Table that contains BioCyc Id, BiGG Id & more
        - model_entities (pd.DataFrame): BiGG Ids of entities in the model 
        - metabolites (bool): True if names of metabolites should be added, otherwise false

    Returns:
        pd.DataFrame: Table containing entities present in KEGG/BioCyc but not in the model
    """
    db = 'KEGG' if 'KEGG' in complete_df.columns else 'BioCyc'  # Find out which database was used
    
    # Get only IDs that are not in model
    mapp = complete_df.set_index('bigg_id')
    entities = model_entities.set_index('bigg_id')
    entities_missing_in_model = mapp[~mapp.index.isin(
        entities.index)].reset_index()
    
    db_ids = entities_missing_in_model.groupby('bigg_id')[db].agg(set)  # Get a set of all BioCyc/KEGG IDs belonging to one BiGG ID
    
    # Add set of BioCyc IDs belonging to one BiGG ID to the dataframe
    entities_missing_in_model.set_index('bigg_id', inplace=True)
    entities_missing_in_model.loc[:, db] = db_ids
    entities_missing_in_model.reset_index(inplace=True)
    
    if 'id_group' in entities_missing_in_model.columns:  # Remove reaction ID duplicates but keep all realted BiGG & BioCyc IDs in a list
        aliases = entities_missing_in_model.groupby(['compartment', 'id_group'])['bigg_id'].agg(set)  # Get a set of the 'duplicated' BiGG reaction IDs -> aliases
        entities_missing_in_model.drop_duplicates(['compartment', 'id_group'], inplace=True, ignore_index=True)  # Drop duplicates where compartments & id_group same
        
        # Add set of BiGG ID aliases to the dataframe
        entities_missing_in_model.set_index(['compartment', 'id_group'], inplace=True)
        entities_missing_in_model.loc[:, 'bigg_aliases'] = aliases
        entities_missing_in_model.reset_index(inplace=True)
        
        entities_missing_in_model.drop(labels='id_group', axis=1, inplace=True)  # id_group is not longer necessary
        
    entities_missing_in_model.drop_duplicates(subset='bigg_id', inplace=True, ignore_index=True)  # Remove BiGG ID duplicates
    
    # Add name column to dataframe
    def get_name_from_bigg(bigg_id: str):
        bigg_db = 'bigg_metabolites' if metabolites else 'bigg_reactions'
        query = f"SELECT name FROM {bigg_db} WHERE bigg_id=\'{bigg_id}\'"
        name_from_bigg = con.execute(query).fetchone()[0]
        return name_from_bigg
    
    con = sqlite3.connect(PATH_TO_DB)  # Open connection to database
    entities_missing_in_model['name'] = entities_missing_in_model['bigg_id'].map(get_name_from_bigg)
    con.close()
    
    # Add compartment ID to all BiGG metabolites that were added due to filtering for BiGG metabolites in BiGG reactions
    if metabolites:
        def get_compartment_from_id(bigg_id: str):
            compartment = bigg_id[-1]
            return compartment if compartment in COMPARTMENTS else np.nan  # To filter the incorrect compartments out
        
        entities_missing_in_model['compartment'] = entities_missing_in_model.apply(
            lambda row: get_compartment_from_id(row['bigg_id']), axis=1)
        entities_missing_in_model.dropna(subset=['compartment'], inplace=True)  # Drop all BiGG metabolite IDs which have no valid compartment
    
    return entities_missing_in_model


def add_stoichiometric_values_to_reacs(missing_reacs: pd.DataFrame) -> pd.DataFrame:
    """Adds for each reaction a dictionary containing the reactants & products as dictionaries with the BiGG Metabolite 
        ID as key & the respective absolute stoichiometric value as value
        
    Args:
        - missing_reacs (pd.DataFrame): Table containing missing reactions (Only requires a column containing BiGG IDs)
            
    Returns:
        pd.DataFrame: Table where for each BiGG reaction ID a dictionary containing reactants & products exists 
    """
    
    def get_reactants_and_products_dicts(reaction_id: str) -> list[dict]:
        reactants = {}
        products = {}
        
        metabs_from_reac = requests.get(BIGG_REACTIONS_URL + reaction_id).json()['metabolites']

        for compound_dict in metabs_from_reac:
            complete_bigg_id = None
            if compound_dict.get('compartment_bigg_id'):
                complete_bigg_id = f"{compound_dict.get('bigg_id')}_{compound_dict.get('compartment_bigg_id')}"
            else:
                complete_bigg_id = compound_dict.get('bigg_id')
            if compound_dict.get('stoichiometry') < 0:
                reactants[complete_bigg_id] = abs(compound_dict.get('stoichiometry'))
            elif compound_dict.get('stoichiometry') > 0:
                products[complete_bigg_id] = abs(compound_dict.get('stoichiometry'))
                
        return str({'reactants': reactants, 'products': products})
                
    missing_reacs['bigg_reaction']= missing_reacs.apply(
        lambda row: get_reactants_and_products_dicts(str(row['bigg_id'])), axis=1)  #, missing_reacs['bigg_products'], result_type='expand'
      
    return missing_reacs
 


#!/usr/bin/env python
"""Provides functions to automate the addition of SBO terms to the model

Script written by Elisabeth Fritze in her bachelor thesis.
Modified by Gwendolyn O. Gusak during her master thesis.
Commented by Famke Bäuerle and extended by Nantia Leonidou.

It is splitted into a lot of small functions which are all annotated, however when using it for SBO-Term annotation it only makes sense to run the "main" function: sbo_annotation(model_libsbml, database_user, database_name) if you want to continue with the model. The smaller functions might be useful if special information is needed for a reaction without the context of a bigger model or when the automated annotation fails for some reason.
"""

import re
import sqlite3
from libsbml import SpeciesReference, Compartment, Reaction, Species
from libsbml import Model as libModel
from refinegems.databases import PATH_TO_DB

__author__ = "Elisabeth Fritze, Gwendolyn O. Gusak, Famke Baeuerle and Nantia Leonidou"


def getCompartmentlessSpeciesId(speciesReference: SpeciesReference) -> str:
    """Determines wheter a species has compartment by its refernece

    Args:
        - speciesReference (SpeciesReference): Reference to species

    Returns:
        libsbml-species-id: id of species without compartment
    """
    speciesId = speciesReference.getSpecies()
    species = speciesReference.getModel().getSpecies(speciesId)
    compartment = species.getCompartment()
    wasteStringLen = len(compartment) + 1
    return(speciesId[:-wasteStringLen])


def getCompartmentFromSpeciesRef(speciesReference: SpeciesReference) -> Compartment:
    """Extracts compartment from a species by its reference

    Args:
        - speciesReference (SpeciesReference): Reference to species

    Returns:
        Compartment: Compartment which the species lives in
    """
    speciesId = speciesReference.getSpecies()
    species = speciesReference.getModel().getSpecies(speciesId)
    compartment = species.getCompartment()
    return compartment


def returnCompartment(id):
    """Helper to split compartment id"""
    return id[-1]


def getReactantIds(reac: Reaction) -> list[str]:
    """Extracts reactants (metabolites) of reaction

    Args:
        - reac (Reaction): Reaction from sbml model

    Returns:
        list[str]: Reactants (metabolites) ids
    """
    list = []
    for metabolite in reac.getListOfReactants():
        list.append(metabolite.getSpecies())
    return list


def getCompartmentlessReactantIds(reac: Reaction):
    """Extracts reactants which have no compartment information

    Args:
        - reac (Reaction): Reaction from sbml model

    Returns:
        list: reactants (metabolites) without compartments
    """
    list = []
    for metabolite in reac.getListOfReactants():
        list.append(getCompartmentlessSpeciesId(metabolite))
    return list


def getProductIds(reac: Reaction):
    """Extracts products (metabolites) of reaction

    Args:
        - reac (Reaction): Reaction from sbml model

    Returns:
        list: products (metabolites) ids
    """
    list = []
    for metabolite in reac.getListOfProducts():
        list.append(metabolite.getSpecies())
    return list


def getCompartmentlessProductIds(reac: Reaction):
    """Extracts products which have no compartment information

    Args:
        - reac (Reaction): Reaction from sbml model

    Returns:
        list: products (metabolites) without compartments
    """
    list = []
    for metabolite in reac.getListOfProducts():
        list.append(getCompartmentlessSpeciesId(metabolite))
    return list


def getListOfMetabolites(reac: Reaction):
    """Extracts list of metabolites of the reaction

    Args:
        - reac (Reaction): Reaction from sbml model

    Returns:
        list: metabolites that are part of the reaction
    """
    list = []
    for reactant in reac.getListOfReactants():
        list.append(reactant)
    for product in reac.getListOfProducts():
        list.append(product)
    return list


def getMetaboliteIds(reac: Reaction):
    """Extracts list of metabolite ids of reaction

    Args:
        - reac (Reaction): Reaction from sbml model

    Returns:
        list: metabolite ids
    """
    return getReactantIds(reac) + getProductIds(reac)


def getCompartmentlessMetaboliteIds(reac: Reaction):
    """Extracts metabolites which have no compartment information

    Args:
        - reac (Reaction): Reaction from sbml model

    Returns:
        list: all metabolites which have no compartment
    """
    return getCompartmentlessReactantIds(
        reac) + getCompartmentlessProductIds(reac)


def getReactantCompartmentList(reac: Reaction):
    """Extracts compartments of reactants

    Args:
        - reac (Reaction): Reaction from sbml model

    Returns:
        set: compartment information of all reactants (metabolites)
    """
    compartments = []
    for metabolite in reac.getListOfReactants():
        compartment = getCompartmentFromSpeciesRef(metabolite)
        compartments.append(compartment)
    return set(compartments)


def getProductCompartmentList(reac: Reaction):
    """Extracts compartments of products

    Args:
        - reac (Reaction): Reaction from sbml model

    Returns:
        set: compartment information of all products (metabolites)
    """
    compartments = []
    for metabolite in reac.getListOfProducts():
        compartment = getCompartmentFromSpeciesRef(metabolite)
        compartments.append(compartment)
    return set(compartments)


def getCompartmentList(reac: Reaction):
    """Extracts compartments of metabolites

    Args:
        - reac (Reaction): Reaction from sbml model

    Returns:
        set: compartment information of all metabolites
    """
    metabolites = getListOfMetabolites(reac)
    compartments = []
    for metabolite in metabolites:
        compartments.append(getCompartmentFromSpeciesRef(metabolite))
    return set(compartments)


def getCompartmentDict(reac: Reaction):
    """sorts metabolites by compartment

    Args:
        - reac (Reaction): Reaction from sbml model

    Returns:
        dict: compartment as key and metabolites as values
    """
    compartmentDict = {}
    for compartment in getCompartmentList(reac):
        compartmentDict[compartment] = []
        for metabolite in getListOfMetabolites(reac):
            if '_' + compartment in metabolite.getSpecies():
                compartmentDict[compartment].append(
                    getCompartmentlessSpeciesId(metabolite))
    return compartmentDict


def moreThanTwoCompartmentTransport(reac: Reaction):
    """check if reaction traverses more than 2 compartments

    Args:
        - reac (Reaction): Reaction from sbml model

    Returns:
        bool: True if reaction traverses more than 2 compartments
    """
    return len(getCompartmentList(reac)) > 2


def isProtonTransport(reac: Reaction):
    """check if reaction is proton transport

    Args:
        - reac (Reaction): Reaction from sbml model

    Returns:
        bool: True if reaction is proton transport
    """
    reactants = reac.getListOfReactants()
    products = reac.getListOfProducts()
    protonTransport = False
    if len(products) == len(reactants) == 1:
        if len(getCompartmentList(reac)) == 2:
            reactant = getCompartmentlessSpeciesId(reactants[0])
            product = getCompartmentlessSpeciesId(products[0])
            protonTransport = reactant == product == 'M_h'
    return protonTransport


def soleProtonTransported(reac: Reaction):
    """check if reaction is transport powered by one H

    Args:
        - reac (Reaction): Reaction from sbml model

    Returns:
        bool: True if reaction is transport powered by one H
    """
    dict = getCompartmentDict(reac)
    soleProton = False
    for compartment in dict:
        if len(dict[compartment]) == 1:
            soleProton = soleProton or dict[compartment][0] == "M_h"
    return soleProton and not isProtonTransport(
        reac) and not moreThanTwoCompartmentTransport(reac)


def getECNums(reac: Reaction):
    """Extracts EC-Code from the reaction annotations

    Args:
        - reac (Reaction): Reaction from sbml model

    Returns:
        list: all EC-Numbers of the reaction
    """
    lines = reac.getAnnotationString().split("\n")
    ECNums = []
    for line in lines:
        if 'ec-code' in line:
            try:
                ECNums.append(line.split('ec-code:')[1][:-3])
            except (IndexError):
                ECNums.append(line.split('ec-code/')[1][:-3])
    return ECNums


def splitTransportBiochem(reac: Reaction):
    """Tests if reaction traverses more than 1 compartment and set SBO Term

    Args:
        - reac (Reaction): Reaction from sbml model
    """
    if len(getCompartmentList(reac)) > 1 and not soleProtonTransported(reac):
        reac.setSBOTerm('SBO:0000655')
    else:
        reac.setSBOTerm('SBO:0000176')


def checkSink(reac: Reaction):
    """Tests if reac is sink and sets SBO Term if true

    Args:
        - reac (Reaction): Reaction from sbml model
    """
    if "_SINK_" in reac.getId() or "_SK_" in reac.getId():
        reac.setSBOTerm('SBO:0000632')


def checkExchange(reac: Reaction):
    """Tests if reac is exchange and sets SBO Term if true

    Args:
        - reac (Reaction): Reaction from sbml model
    """
    if "_EX_" in reac.getId():
        reac.setSBOTerm('SBO:0000627')


def checkDemand(reac: Reaction):
    """Tests if reac is demand and sets SBO Term if true

    Args:
        - reac (Reaction): Reaction from sbml model
    """
    if "_DM_" in reac.getId():
        reac.setSBOTerm('SBO:0000628')


def checkBiomass(reac: Reaction):  # memote says growth is biomass
    """Tests if reac is biomass / growth and sets SBO Term if true

    Args:
        - reac (Reaction): Reaction from sbml model
    """
    # Use regex to generalise check for growth/biomass reaction
    regex = 'growth|_*biomass\d*_*'
    if re.match(regex, reac.getId(), re.IGNORECASE):
        reac.setSBOTerm('SBO:0000629')


def checkPassiveTransport(reac: Reaction):
    """Tests if reac is passive transport and sets SBO Term if true

    Args:
        - reac (Reaction): Reaction from sbml model
    """
    reactants = reac.getListOfReactants()
    products = reac.getListOfProducts()
    if len(reactants) == len(products) == 1:
        reactant = reactants[0].getSpecies()
        product = products[0].getSpecies()
        if returnCompartment(reactant) != returnCompartment(product):
            reac.setSBOTerm('SBO:0000658')


def checkActiveTransport(reac: Reaction):
    """Tests if reac is active transport (uses atp/pep) and sets SBO Term if true

    Args:
        - reac (Reaction): Reaction from sbml model
    """
    reactantIds = []
    for metabolite in reac.getListOfReactants():
        reactantIds.append(metabolite.getSpecies())
    if 'M_atp_c' in reactantIds or 'M_pep_c' in reactantIds:
        reac.setSBOTerm('SBO:0000657')
        if reac.getReversible():
            print("Error, active reaction but reversible " + reac.getId())


def checkCoTransport(reac: Reaction):
    """Tests if reac is co-transport and sets SBO Term if true

    Args:
        - reac (Reaction): Reaction from sbml model
    """
    reactants = reac.getListOfReactants()
    if len(reactants) > 1:
        reac.setSBOTerm('SBO:0000654')


def splitSymAntiPorter(reac: Reaction):
    """Tests if reac is sym- or antiporter and sets SBO Term if true

    Args:
        - reac (Reaction): Reaction from sbml model
    """
    if len(getCompartmentList(reac)) > 2:
        pass
    elif 1 == len(getReactantCompartmentList(reac)) == len(getProductCompartmentList(reac)):
        reac.setSBOTerm('SBO:0000659')
    else:
        reac.setSBOTerm('SBO:0000660')


def checkPhosphorylation(reac: Reaction):
    """Tests if reac is phosphorylase / kinase and sets SBO Term if true

    Args:
        - reac (Reaction): Reaction from sbml model
    """
    name = reac.getName()
    atpIsReactant = 'M_atp_c' in getReactantIds(reac)
    adpIsProduct = 'M_adp_c' in getProductIds(reac)
    if ' phosphorylase' in name or 'kinase' in name:
        reac.setSBOTerm('SBO:0000216')


def hasReactantPair(reac: Reaction, met1: Species, met2: Species) -> bool:
    """| Checks if a pair of metabolites is present in reaction
       | needed for special reactions like redox or deamination

    Args:
        - reac (Reaction): Reaction from sbml model
        - met1 (Species): metabolite 1 of metabolite pair
        - met2 (Species): metabolite 2 of metabolite pair

    Returns:
        bool: True if one of the metabolites is in reactants and the other in products
    """
    reactants = getCompartmentlessReactantIds(reac)
    products = getCompartmentlessProductIds(reac)
    return (met1 in reactants and met2 in products) or (
        met2 in reactants and met1 in products)


def checkRedox(reac: Reaction):
    """Tests if reac is redox and sets SBO Term if true

    Args:
        - reac (Reaction): Reaction from sbml model
    """
    isRedox = False
    isRedox = isRedox or hasReactantPair(reac, 'M_pyr', 'M_lac_L')
    isRedox = isRedox or hasReactantPair(reac, 'M_pyr', 'M_lac_D')
    isRedox = isRedox or hasReactantPair(reac, 'M_nad', 'M_nadh')
    isRedox = isRedox or hasReactantPair(reac, 'M_nadp', 'M_nadph')
    isRedox = isRedox or hasReactantPair(reac, 'M_fad', 'M_fadh2')
    isRedox = isRedox or hasReactantPair(reac, 'M_h20', 'M_h2o2')
    if isRedox:
        reac.setSBOTerm('SBO:0000200')


def checkGlycosylation(reac: Reaction):
    """Tests if reac is glycosylation and sets SBO Term if true

    Args:
        - reac (Reaction): Reaction from sbml model
    """
    isGlycosylation = False
    isGlycosylation = isGlycosylation or hasReactantPair(
        reac, 'M_ppi', 'M_prpp')
    isGlycosylation = isGlycosylation or hasReactantPair(
        reac, 'M_udpglcur', 'M_udp')
    if isGlycosylation:
        reac.setSBOTerm('SBO:0000217')


def checkDecarbonylation(reac: Reaction):
    """Tests if reac is decarbonylation and sets SBO Term if true

    Args:
        - reac (Reaction): Reaction from sbml model
    """
    if not reac.getReversible():
        if 'M_co' in getCompartmentlessProductIds(reac):
            reac.setSBOTerm('SBO:0000400')


def checkDecarboxylation(reac: Reaction):
    """Tests if reac is decarboxylation and sets SBO Term if true

    Args:
        - reac (Reaction): Reaction from sbml model
    """
    if not reac.getReversible():
        if 'M_co2' in getCompartmentlessProductIds(reac):
            reac.setSBOTerm('SBO:0000399')


def checkDeamination(reac: Reaction):
    """Tests if reac is deamination and sets SBO Term if true

    Args:
        - reac (Reaction): Reaction from sbml model
    """
    if not reac.getReversible():
        waterAdded = 'M_h2o' in getCompartmentlessReactantIds(reac)
        nh4Removed = 'M_nh4' in getCompartmentlessProductIds(reac)
        if waterAdded and nh4Removed:
            reac.setSBOTerm('SBO:0000401')


def checkRedoxViaEC(reac: Reaction):
    """Tests if reac is redox by its EC-Code and sets SBO Term if true

    Args:
        - reac (Reaction): Reaction from sbml model
    """
    if len(getECNums(reac)) == 1:
        if getECNums(reac)[0].startswith('1'):
            reac.setSBOTerm('SBO:0000200')


def checkAcetylationViaEC(reac: Reaction):
    """Tests if reac is acetylation by its EC-Code and sets SBO Term if true

    Args:
        - reac (Reaction): Reaction from sbml model
    """
    if len(getECNums(reac)) == 1:
        if getECNums(reac)[0].startswith('2.3.1'):
            reac.setSBOTerm('SBO:0000215')


def checkGlycosylationViaEC(reac: Reaction):
    """Tests if reac is glycosylation by its EC-Code and sets SBO Term if true

    Args:
        - reac (Reaction): Reaction from sbml model
    """
    if len(getECNums(reac)) == 1:
        if getECNums(reac)[0].startswith('2.4'):
            reac.setSBOTerm('SBO:0000217')


def checkMethylationViaEC(reac: Reaction):
    """Tests if reac is methylation by its EC-Code and sets SBO Term if true

    Args:
        - reac (Reaction): Reaction from sbml model
    """
    if len(getECNums(reac)) == 1:
        if getECNums(reac)[0].startswith('2.1.1'):
            reac.setSBOTerm('SBO:0000214')


def checkTransaminationViaEC(reac: Reaction):
    """Tests if reac is transamination by its EC-Code and sets SBO Term if true

    Args:
        - reac (Reaction): Reaction from sbml model
    """
    if len(getECNums(reac)) == 1:
        if getECNums(reac)[0].startswith('2.6.1'):
            reac.setSBOTerm('SBO:0000403')


def checkDeaminationViaEC(reac: Reaction):
    """Tests if reac is deamination by its EC-Code and sets SBO Term if true

    Args:
        - reac (Reaction): Reaction from sbml model
    """
    if len(getECNums(reac)) == 1:
        if getECNums(reac)[0].startswith('3.5.4'):
            reac.setSBOTerm('SBO:0000401')


def checkDecarboxylationViaEC(reac: Reaction):
    """Tests if reac is decarboxylation by its EC-Code and sets SBO Term if true

    Args:
        - reac (Reaction): Reaction from sbml model
    """
    if len(getECNums(reac)) == 1:
        if getECNums(reac)[0].startswith('4.1.1'):
            reac.setSBOTerm('SBO:0000399')


def checkIsomerisationViaEC(reac: Reaction):
    """Tests if reac is isomerisation by its EC-Code and sets SBO Term if true

    Args:
        - reac (Reaction): Reaction from sbml model
    """
    if len(getECNums(reac)) == 1:
        if getECNums(reac)[0].startswith('5'):
            reac.setSBOTerm('SBO:0000377')


def checkHydrolysisViaEC(reac: Reaction):
    """Tests if reac is hydrolysis by its EC-Code and sets SBO Term if true

    Args:
        - reac (Reaction): Reaction from sbml model
    """
    if len(getECNums(reac)) == 1:
        if getECNums(reac)[0].startswith('3'):
            reac.setSBOTerm('SBO:0000376')


def addSBOviaEC(reac: Reaction, cur):
    """Adds SBO terms based on EC numbers given in the annotations of a reactions

    Args:
        - reac (Reaction): Reaction from sbml model
        - cur (sqlite3.connect.cursor): Used to access the sqlite3 database
    """
    if len(getECNums(reac)) == 1:
        ECnum = getECNums(reac)[0]
        splittedEC = ECnum.split(".")
        if len(splittedEC) == 4:
            ECpos1 = splittedEC[0]
            ECpos1to2 = ECpos1 + "." + splittedEC[1]
            ECpos1to3 = ECpos1to2 + "." + splittedEC[2]
            query4 = cur.execute("""SELECT sbo_term
                                     FROM ec_to_sbo 
                                    WHERE ecnum = ?""", [ECnum])
            result4 = cur.fetchone()
            if result4 is not None:
                sbo4 = result4[0]
                reac.setSBOTerm(sbo4)
            else:
                query3 = cur.execute("""SELECT sbo_term
                                        FROM ec_to_sbo 
                                        WHERE ecnum = ?""", [ECpos1to3])
                result3 = cur.fetchone()
                if result3 is not None:
                    sbo3 = result3[0]
                    reac.setSBOTerm(sbo3)
                else:
                    query2 = cur.execute("""SELECT sbo_term
                                            FROM ec_to_sbo t
                                            WHERE ecnum = ?""", [ECpos1to2])
                    result2 = cur.fetchone()
                    if result2 is not None:
                        sbo2 = result2[0]
                        reac.setSBOTerm(sbo2)
                    else:
                        query1 = cur.execute("""SELECT sbo_term
                                                FROM ec_to_sbo
                                                WHERE ecnum = ?""", [ECpos1])
                        result1 = cur.fetchone()
                        if result1 is not None:
                            sbo1 = result1[0]
                            reac.setSBOTerm(sbo1)


def addSBOfromDB(reac: Reaction, cur) -> bool:
    """Adds SBO term based on bigg id of a reaction

    Args:
        - reac (Reaction): Reaction from sbml model
        - cur (sqlite3.connect.cursor): Used to access the sqlite3 database

    Returns:
        bool: True if SBO Term was changed
    """
    reacid = reac.getId()
    query = cur.execute("""SELECT sbo_term 
                           FROM bigg_to_sbo 
                           WHERE bigg_reactionid = ?""", [reacid])
    result = cur.fetchone()
    if result is not None:
        sbo_term = result[0]
        reac.setSBOTerm(sbo_term)
        return True
    else:
        return False

### functions below from Nantia ###

def addSBOforMetabolites(model):
    # add metabolites SBO
    for met in model.species:
        met_id = met.getId()
        model.getSpecies(met_id).setSBOTerm("SBO:0000247")


def addSBOforGenes(model):
    # add genes SBO
    model_fbc = model.getPlugin("fbc")
    # if model has genes
    if model_fbc is not None:
        for gene in model_fbc.getListOfGeneProducts():
            gene.setSBOTerm("SBO:0000243")


def addSBOforModel(model):
    model.setSBOTerm("SBO:0000624")


def addSBOforGroups(model):
    mplugin = model.getPlugin("groups")
    # if groups are in model defined
    if mplugin is not None:
        for grp in mplugin.getListOfGroups():
            grp.setSBOTerm("SBO:0000633")


def addSBOforParameters(model):
    for param in model.getListOfParameters():
        # reaction bounds
        if 'R_' in param.getId():
            param.setSBOTerm("SBO:0000625")
        # default set bounds
        else:
            param.setSBOTerm("SBO:0000626")


def addSBOforCompartments(model):
    for cmp in model.getListOfCompartments():
        cmp.setSBOTerm("SBO:0000290")   

### end functions from Nantia ###

def sbo_annotation(model_libsbml: libModel) -> libModel:
    """Executes all steps to annotate SBO terms to a given model (former main function of original script by Elisabeth Fritze)

    Args:
        - model_libsbml (libModel): Model loaded with libsbml
      
    Returns:
        libModel: Modified model with SBO terms
    """
    open_con = sqlite3.connect(PATH_TO_DB)
    open_cur = open_con.cursor()

    for reaction in model_libsbml.reactions:
        if not addSBOfromDB(reaction, open_cur):
            reaction.unsetSBOTerm()
            splitTransportBiochem(reaction)
            checkBiomass(reaction)
            checkSink(reaction)
            checkExchange(reaction)
            checkDemand(reaction)
            if reaction.getSBOTermID() == 'SBO:0000655': #transporter
                checkPassiveTransport(reaction)
                checkActiveTransport(reaction)
                if reaction.getSBOTermID() != 'SBO:0000657':
                    checkCoTransport(reaction)
                    if reaction.getSBOTermID() == 'SBO:0000654':
                        splitSymAntiPorter(reaction)
            if reaction.getSBOTermID() == 'SBO:0000176': #metabolic reaction
                addSBOviaEC(reaction, open_cur)
            if reaction.getSBOTermID() == 'SBO:0000176':
                checkRedox(reaction)
                checkGlycosylation(reaction)
                checkDecarbonylation(reaction)
                checkDecarboxylation(reaction)
                checkDeamination(reaction)
                checkPhosphorylation(reaction)
    
    ### functions from Nantia ###            
    # addSBOforMetabolites(model_libsbml)
    # addSBOforGenes(model_libsbml)
    # addSBOforModel(model_libsbml)
    # addSBOforGroups(model_libsbml)
    # addSBOforParameters(model_libsbml)
    # addSBOforCompartments(model_libsbml)
    
    open_cur.close()
    open_con.close()
    return model_libsbml


#!/usr/bin/env python
"""Can be used to polish a model (created with CarveMe v.1.5.1)

The newer version of CarveMe leads to some irritations in the model, these scripts enable for example the addition of BiGG Ids to the annotations as well as a correct formatting of the annotations.
"""

import re, logging
from libsbml import Model as libModel
from libsbml import Species, Reaction, Unit, UnitDefinition, SBase, UNIT_KIND_MOLE, UNIT_KIND_GRAM, UNIT_KIND_LITRE, UNIT_KIND_SECOND, MODEL_QUALIFIER, BQM_IS, BQM_IS_DERIVED_FROM, BQM_IS_DESCRIBED_BY, BIOLOGICAL_QUALIFIER, BQB_IS, BQB_IS_HOMOLOG_TO, BiolQualifierType_toString, ModelQualifierType_toString
from Bio import Entrez
from tqdm.auto import tqdm
from sortedcontainers import SortedDict, SortedSet
from refinegems.cvterms import add_cv_term_units, add_cv_term_metabolites, add_cv_term_reactions, add_cv_term_genes, generate_cvterm, metabol_db_dict, reaction_db_dict, MIRIAM, OLD_MIRIAM
from refinegems.io import search_ncbi_for_gpr, parse_fasta_headers
from colorama import init as colorama_init
from colorama import Fore, Style

__author__ = "Famke Baeuerle and Gwendolyn O. Gusak"
    
 
#----------- Functions to add URIs from the entity IDs to the annotation field for metabolites & reactions ------------#       
def add_metab(entity_list: list[Species], id_db: str):
    """| Adds the ID of metabolites as URI to the annotation field
       | For a VMH model, additionally, the corresponding BiGG IDs are added! 
	   | (Currently, only BiGG & VMH IDs supported!)

    Args:
        - entity_list (list): libSBML ListOfSpecies
        - id_db (str): Name of the database of the IDs contained in a model 
    """
    vmh_cut_pattern = '__\d+_' # To extract BiGG identifier
    
    for entity in entity_list:
        
        # Get ID and remove 'M_'
        current_id = entity.getId()
        current_id = current_id[2:]
        
        # Unset annotations if no CV terms exist
        if entity.getNumCVTerms() == 0:
            entity.unsetAnnotation()
        
        # If database 'VMH' was specified, extract BiGG ID with vmh_cut_pattern
        if id_db == 'VMH':
            current_id_cut = re.split(vmh_cut_pattern, current_id)
            current_id = ''.join(current_id_cut[:2])

        # Use current_id as metaid if no metaid is present   
        if not entity.isSetMetaId():
            entity.setMetaId(f'meta_M_{current_id}')
        
        # Remove compartment specifier from ID for annotation   
        id_for_anno = current_id[:-2]

		# Add ID as URI to annotation
        add_cv_term_metabolites(id_for_anno, id_db, entity)
        
        if id_db == 'VMH':
            # Add BiGG ID to annotation, additionally
            add_cv_term_metabolites(id_for_anno, 'BIGG', entity)
            
           
def add_reac(entity_list: list[Reaction], id_db: str):
    """| Adds the ID of reactions as URI to the annotation field
       | (Currently, only BiGG & VMH IDs supported!)

        Args:
            - entity_list (list): libSBML ListOfReactions
            - id_db (str): Name of the database of the IDs contained in a model             
    """
    # Use regex to generalise check for growth/biomass reaction
    regex = 'growth|_*biomass\d*_*'
    
    for entity in entity_list:
        
        # Get ID and remove 'R_'
        current_id = entity.getId()
        if current_id[:2] == 'R_':
            current_id = current_id[2:]
        
        # Readjusted to fit to the metaid pattern from the metabolites      
        if not entity.isSetMetaId():
            entity.setMetaId(f'meta_R_{current_id}')

        if not re.fullmatch(regex, current_id, re.IGNORECASE):
            
            # Unset annotations if no CV terms exist
            if entity.getNumCVTerms() == 0:
                entity.unsetAnnotation()
            
            # Add ID as URI to annotation   
            add_cv_term_reactions(current_id, id_db, entity)


#----------- Functions to transfer URIs from the notes field to the annotations for metabolites & reactions -----------# 
def cv_notes_metab(species_list: list[Species]):
    """| Checks the notes field for information which should be in the annotation field
       | removes entry from notes and adds it as URL to the CVTerms of a metabolite

    Args:
        - species_list (list): libSBML ListOfSpecies
    """

    for species in species_list:
        if not species.isSetMetaId():
            species.setMetaId('meta_' + species.getId())
        notes_list = []
        elem_used = []
        notes_string = species.getNotesString().split('\n')
        for elem in notes_string:
            for db in metabol_db_dict.keys():
                if '<p>' + db in elem:
                    elem_used.append(elem)
                    #print(elem.strip()[:-4].split(': ')[1])
                    fill_in = re.split(':\s*', elem.strip()[:-4])[1]
                    if (';') in fill_in and db != 'INCHI':
                        entries = fill_in.split(';')
                        for entry in entries:
                            add_cv_term_metabolites(entry.strip(), db, species)
                    else:
                        add_cv_term_metabolites(fill_in, db, species)

        for elem in notes_string:
            if elem not in elem_used and elem not in notes_list:
                notes_list.append(elem)

        # Adding new, shortened notes
        new_notes = ' '.join([str(elem) + '\n' for elem in notes_list])
        species.unsetNotes()
        species.setNotes(new_notes)
        #print(species.getAnnotationString())


def cv_notes_reac(reaction_list: list[Reaction]):
    """| Checks the notes field for information which should be in the annotation field
       | removes entry from notes and adds it as URL to the CVTerms of a reaction

    Args:
        - reaction_list (list): libSBML ListOfReactions
    """

    for reaction in reaction_list:
        if not reaction.isSetMetaId():
            reaction.setMetaId('meta_' + reaction.getId())
        notes_list = []
        elem_used = []
        notes_string = reaction.getNotesString().split('\n')

        for elem in notes_string:
            for db in reaction_db_dict.keys():
                if '<p>' + db in elem:
                    elem_used.append(elem)
                    #print(elem.strip()[:-4].split(': ')[1])
                    fill_in = re.split(':\s*', elem.strip()[:-4])[1]
                    if (';') in fill_in:
                        entries = fill_in.split(';')
                        for entry in entries:
                            add_cv_term_reactions(entry.strip(), db, reaction)
                    else:
                        add_cv_term_reactions(fill_in, db, reaction)

        for elem in notes_string:
            if elem not in elem_used and elem not in notes_list:
                notes_list.append(elem)

        # Adding new, shortened notes
        new_notes = ' '.join([str(elem) + '\n' for elem in notes_list])
        reaction.unsetNotes()
        reaction.setNotes(new_notes)


#--------------- Function to set boundary condition & constant for metabolites & reactions ----------------------------# 
def polish_entities(entity_list: list, metabolite: bool):
    """Sets boundary condition and constant if not set for a metabolite

    Args:
        - entity_list (list): libSBML ListOfSpecies or ListOfReactions
        - metabolite (boolean): flag to determine whether entity = metabolite
    """
    for entity in entity_list:
        if metabolite:  # some polishing
            if not entity.getBoundaryCondition():
                entity.setBoundaryCondition(False)
            if not entity.getConstant():
                entity.setConstant(False) 


#------------------------------------ Functions to create units & UnitDefinitions -------------------------------------# 
def create_unit(
    model_specs: tuple[int], meta_id: str, kind: str, e: int, m: int, s: int, uri_is: str='', uri_idf: str=''
    ) -> Unit:
    """Creates unit for SBML model according to arguments

    Args:
        - model_specs (tuple):    Level & Version of SBML model
        - meta_id (str):          Meta ID for unit (Neccessary for URI)
        - kind (str):             Unit kind constant (see libSBML for available constants)
        - e (int):                Exponent of unit
        - m (int):                Multiplier of unit
        - s (int):                Scale of unit 
        - uri_is (str):           URI supporting the specified unit
        - uri_idf (str):          URI supporting the derived from unit
      
    Returns:
        Unit: libSBML unit object
    """
    unit = Unit(*model_specs)
    unit.setKind(kind)
    unit.setMetaId(f'meta_{meta_id}_{unit.getKind()}')
    unit.setExponent(e)
    unit.setMultiplier(m)
    unit.setScale(s)
    if uri_is:
        add_cv_term_units(uri_is, unit, BQM_IS)
    if uri_idf:
        add_cv_term_units(uri_idf, unit, BQM_IS_DERIVED_FROM)
    return unit


def create_unit_definition(model_specs: tuple[int], identifier: str, name: str, 
                           units: list[Unit]) -> UnitDefinition:
    """Creates unit definition for SBML model according to arguments
   
    Args:
        - model_specs (tuple): Level & Version of SBML model
        - identifier (str):    Identifier for the defined unit
        - name (str):          Full name of the defined unit
        - units (list):        All units the defined unit consists of
         
    Returns:
        UnitDefinition: libSBML unit definition object
    """
    unit_definition = UnitDefinition(*model_specs)
    unit_definition.setId(identifier)
    unit_definition.setMetaId(f'meta_{identifier}')
    unit_definition.setName(name)
      
    # Iterate over all units provided for the unit definition & add the units
    for unit in units:
        unit_definition.addUnit(unit)
   
    return unit_definition


#----------------------------------- Function to create FBA unit & UnitDefinitions ------------------------------------#
def create_fba_units(model: libModel) -> list[UnitDefinition]:
    """Creates all fba units required for a constraint-based model
   
    Args:
        - model (libModel): Model loaded with libSBML
         
    Returns:
        list: List of libSBML UnitDefinitions
    """
    # Get model level & version for unit & unit definition
    model_specs = model.getLevel(), model.getVersion()
    
    # Create required units
    litre = create_unit(model_specs, 'litre', UNIT_KIND_LITRE, e=1, m=1, s=-3, uri_is='0000104', uri_idf='0000099')
    mole = create_unit(model_specs, 'mole_0', UNIT_KIND_MOLE, e=1, m=1, s=-3, uri_is='0000040', uri_idf='0000013')
    gram = create_unit(model_specs, 'per_gram_0', UNIT_KIND_GRAM, e=-1, m=1, s=0, uri_is='0000021')
    second = create_unit(model_specs, 'second_0', UNIT_KIND_SECOND, e=1, m=3600, s=0, uri_is='0000032', uri_idf='0000010')
    
    # Create unit definitions for hour & femto litre
    hour = create_unit_definition(
        model_specs, identifier='h', name='Hour', 
        units=[second]
    )
    femto_litre = create_unit_definition(
        model_specs, identifier='fL', name='Femto litres', 
        units=[litre]
    )
    
    # Create unit definitions for millimoles per gram dry weight (mmgdw) & mmgdw per hour
    mmgdw = create_unit_definition(
        model_specs, identifier='mmol_per_gDW', name='Millimoles per gram (dry weight)',
        units=[mole, gram]
    )

    # Create new units mole & gram to get new meta IDs 
    mole = create_unit(model_specs, 'mole_1', UNIT_KIND_MOLE, e=1, m=1, s=-3, uri_is='0000040', uri_idf='0000013')
    gram = create_unit(model_specs, 'per_gram_1', UNIT_KIND_GRAM, e=-1, m=1, s=0, uri_is='0000021')
    # Create new unit second to fit to per hour & get new meta ID
    second = create_unit(model_specs, 'second_1', UNIT_KIND_SECOND, e=-1, m=3600, s=0, uri_is='0000032', uri_idf='0000010')

    mmgdwh = create_unit_definition(
        model_specs, identifier='mmol_per_gDW_per_h', name='Millimoles per gram (dry weight) per hour',
        units=[mole, gram, second]
    )
    add_cv_term_units('pubmed:7986045', mmgdwh, BQM_IS_DESCRIBED_BY)

    return [mmgdwh, mmgdw, hour, femto_litre]


#--------------------------- Functions to print UnitDefinitions with the respective units -----------------------------#
def print_UnitDefinitions(contained_unit_defs: list[UnitDefinition]):
    """Prints a list of libSBML UnitDefinitions as XMLNodes
   
    Args:
        - contained_unit_defs (list): List of libSBML UnitDefinition objects
    """
    for unit_def in contained_unit_defs:
        logging.info(unit_def.toXMLNode())


def print_remaining_UnitDefinitions(model: libModel, list_of_fba_units: list[UnitDefinition]):
    """Prints UnitDefinitions from the model that were removed as these were not contained in the list_of_fba_units

    Args:
        - model (libModel): Model loaded with libSBML
        - list_of_fba_units (list):  List of libSBML UnitDefinitions  
    """
       
    # Get all units already present in the model
    contained_unit_defs = [unit for unit in model.getListOfUnitDefinitions()]
         
    # Check if contained unit fits to one of the created fba units
    for unit_def in list_of_fba_units:
        for contained_unit_def in contained_unit_defs:
         
            current_id = contained_unit_def.getId()
               
            if UnitDefinition.areIdentical(unit_def, contained_unit_def):
                contained_unit_defs.remove(contained_unit_def)
                model.removeUnitDefinition(current_id)
   
    # Only print list if it contains UnitDefinitions         
    if contained_unit_defs:
        logging.info('''
        The following UnitDefinition objects were removed. 
        The reasoning is that
        \t(a) these UnitDefinitions are not contained in the UnitDefinition list of this program and
        \t(b) the UnitDefinitions defined within this program are handled as ground truth.
        Thus, the following UnitDefinitions are not seen as relevant for the model.
        ''')
        print_UnitDefinitions(contained_unit_defs)


#-------------------------------------- Function to add units & UnitDefinitions ---------------------------------------#
def add_fba_units(model: libModel):
    """| Adds:
       |     - mmol per gDW per h
       |     - mmol per gDW 
       |     - hour (h)
       |     - femto litre (fL)
       |
       | to the list of unit definitions (needed for FBA)

    Args:
        - model (libModel): Model loaded with libSBML
    """
    list_of_fba_units = create_fba_units(model)
    
    # If list of unit definitions is not empty, replace all units identical to list_of_fba_units
    # & Print the remaining unit definitions
    if model.getListOfUnitDefinitions():
        print_remaining_UnitDefinitions(model, list_of_fba_units)
    
    for unit_def in list_of_fba_units:
        model.getListOfUnitDefinitions().append(unit_def)
          

def set_default_units(model: libModel):
    """Sets default units of model

    Args:
        - model (libModel): Model loaded with libSBML
    """ 
    for unit in model.getListOfUnitDefinitions():
      
        unit_id = unit.getId()
      
        if re.fullmatch('mmol_per_gDW', unit_id, re.IGNORECASE):
         
            if not (model.isSetExtentUnits() and model.getExtentUnits() == unit_id):
                model.setExtentUnits(unit_id)
            
            if not (model.isSetSubstanceUnits() and model.getSubstanceUnits() == unit_id):
                model.setSubstanceUnits(unit_id)
         
        if not (model.isSetTimeUnits() and model.getTimeUnits() == unit_id) and re.fullmatch('hr?', unit_id, re.IGNORECASE):
            model.setTimeUnits(unit_id)
         
        if not (model.isSetVolumeUnits() and model.getVolumeUnits() == unit_id) and re.fullmatch('fL', unit_id, re.IGNORECASE):
            model.setVolumeUnits(unit_id)


#--------------------------------------- Function to set units fro parameters -----------------------------------------#
def set_units(model: libModel):
    """Sets units of parameters in model

    Args:
        - model (libModel): Model loaded with libSBML
    """
    for param in model.getListOfParameters(): # needs to be added to list of unit definitions aswell
        if any(
            (unit_id := re.fullmatch('mmol_per_gDW_per_hr?', unit.getId(), re.IGNORECASE)) 
            for unit in model.getListOfUnitDefinitions()
            ):
            if not (param.isSetUnits() and param.getUnits() == unit_id.group(0)):
                param.setUnits(unit_id.group(0))
            

#-------------------------- Functions to add default settings for compartments & metabolites --------------------------#            
def add_compartment_structure_specs(model: libModel):
    """| Adds the required specifications for the compartment structure
       | if not set (size & spatial dimension)
        
    Args:
        - model (libModel): Model loaded with libSBML
    """ 
    for compartment in model.getListOfCompartments():
      
        if not compartment.isSetSize():
            compartment.setSize(float('NaN'))
         
        if not compartment.isSetSpatialDimensions():
            compartment.setSpatialDimensions(3)
         
        if any(
            (unit_id := re.fullmatch('fL', unit.getId(), re.IGNORECASE)) for unit in model.getListOfUnitDefinitions()
            ):
            if not (compartment.isSetUnits() and compartment.getUnits() == unit_id.group(0)):
                compartment.setUnits(unit_id.group(0))
         
         
def set_initial_amount(model: libModel):
    """Sets initial amount to all metabolites if not already set or if initial concentration is not set
    
    Args:
        - model (libModel): Model loaded with libSBML
    """
    for species in model.getListOfSpecies():
      
      if not (species.isSetInitialAmount() or species.isSetInitialConcentration()):
         species.setInitialAmount(float('NaN'))
         

#--------------------------------- Function to add URIs from the IDs for GeneProducts ---------------------------------# 
def cv_ncbiprotein(gene_list, email, protein_fasta: str, lab_strain: bool=False):
    """Adds NCBI Id to genes as annotation

    Args:
        - gene_list (list): libSBML ListOfGenes
        - email (str): User Email to access the Entrez database
        - protein_fasta (str): The path to the CarveMe protein.fasta input file
        - lab_strain (bool): Needs to be set to True if strain was self-annotated
                           and/or the locus tags in the CarveMe input file should be kept   
    """
    Entrez.email = email
                    
    id2locus_name = None  # Needs to be initialised, otherwise UnboundLocalError: local variable 'id2locus_name' referenced before assignment          
    if (protein_fasta is not None) and protein_fasta.strip() != '': 
       id2locus_name = parse_fasta_headers(protein_fasta)
       id2locus_name.set_index('protein_id')
    
    genes_missing_annotation = []

    print('Setting CVTerms and removing notes for all genes:')
    for gene in tqdm(gene_list):
        if not gene.isSetMetaId():
            gene.setMetaId('meta_' + gene.getId())
        
        if (gene.getId()[2] == 'W'): #addition to work with KC-Na-01
            entry = gene.getId()[2:-2]
            add_cv_term_genes(entry, 'NCBI', gene, lab_strain)
            name, locus = search_ncbi_for_gpr(entry)
            gene.setName(name)
            gene.setLabel(locus)
        
        elif (gene.getId() != 'G_spontaneous'): # Has to be omitted as no additional data can be retrieved neither from NCBI nor the CarveMe input file
            if 'prot_' in gene.getId():
                id_string = gene.getId().split('prot_')[1].split('_')  # All NCBI CDS protein FASTA files have the NCBI protein identifier after 'prot_' in the FASTA identifier
                ncbi_id = id_string[0]  # If identifier contains no '_', this is full identifier
                
                if (len(id_string) > 2):  # Identifier contains '_'
                # Check that the second entry consists of a sequence of numbers -> Valid RefSeq identifier! 
                # (Needs to be changed if there are other gene idenitfiers used that could contain '_' & need to be handled differently)
                    if re.fullmatch('^\d+\d+$', id_string[1], re.IGNORECASE):
                        ncbi_id = '_'.join(id_string[:2])  # Merge the first two parts with '_' as this is complete identifier
                
                # If identifier matches RefSeq ID pattern   
                if re.fullmatch('^(((AC|AP|NC|NG|NM|NP|NR|NT|NW|WP|XM|XP|XR|YP|ZP)_\d+)|(NZ_[A-Z]{2,4}\d+))(\.\d+)?$', ncbi_id, re.IGNORECASE):
                    add_cv_term_genes(ncbi_id, 'REFSEQ', gene, lab_strain)
                    name, locus = search_ncbi_for_gpr(ncbi_id)
            
                # If identifier only contains numbers 
                # -> Get the corresponding data from the CarveMe input file
                elif re.fullmatch('^\d+$', ncbi_id, re.IGNORECASE):
                    if id2locus_name is not None:
                        name, locus = id2locus_name[id2locus_name['protein_id']==ncbi_id][['name', 'locus_tag']].values[0]
                    else: 
                        genes_missing_annotation.append(ncbi_id)
            
                # If identifier matches ncbiprotein ID pattern
                elif re.fullmatch('^(\w+\d+(\.\d+)?)|(NP_\d+)$', ncbi_id, re.IGNORECASE):
                    add_cv_term_genes(ncbi_id, 'NCBI', gene, lab_strain)
                    name, locus = search_ncbi_for_gpr(ncbi_id)
                
                # Catch all remaining cases that have no valid ID   
                else: 
                    genes_missing_annotation.append(ncbi_id)
            
                # For lab strains use the locus tag from the annotation file   
                if lab_strain and id2locus_name is not None:
                    locus = id2locus_name[id2locus_name['protein_id']==ncbi_id][['locus_tag']].values[0]
            
                if ncbi_id not in genes_missing_annotation:      
                    gene.setName(name)
                    gene.setLabel(locus)
            
        gene.unsetNotes()
    if genes_missing_annotation:    
        logging.info(f'The following {len(genes_missing_annotation)} genes have no annotation, name & label (locus tag): {genes_missing_annotation}')


#---------------- Functions to change the CURIE pattern/CVTerm qualifier & qualifier type ----------------------# 
def get_set_of_curies(curie_list: list[str]) -> SortedDict[str: SortedSet[str]]:
    """| Gets a list of CURIEs
       | & maps the database prefixes to their respective identifier sets
        
    Args:
        - curie_list (list[str]): List containing CURIEs
            
    Returns:
        SortedDict: Sorted dictionary mapping database prefixes from the provided CURIEs to their respective identifier sets also provided by the CURIEs
    """
    curie_dict = SortedDict()
    
    for curie in curie_list:
        
        # Extracts the prefix & identifier part
        if MIRIAM in curie:
            extracted_curie = curie.split(MIRIAM)[1]
        else:
            extracted_curie = curie.split(OLD_MIRIAM)[1]
        
        # Get CURIEs irrespective of pattern
        if '/' in extracted_curie:
            extracted_curie = extracted_curie.split('/')
            
            # Check for certain special cases
            if re.fullmatch('^inchi$', extracted_curie[0], re.IGNORECASE):  # Check for inchi as splitting by '/' splits too much
                prefix = extracted_curie[0].lower()
                identifier = '/'.join(extracted_curie[1:len(extracted_curie)])
            elif re.fullmatch('^brenda$', extracted_curie[0], re.IGNORECASE):
                prefix = 'ec-code'
                identifier = extracted_curie[1]
            elif re.fullmatch('^biocyc$', extracted_curie[0], re.IGNORECASE) or ('metacyc.' in extracted_curie[0]):  # Check for bio- & metacyc
                prefix = 'biocyc'
                identifier = extracted_curie[1].replace('META:', '')
                
                if not curie_dict or (prefix not in curie_dict):
                    curie_dict[prefix] = SortedSet()
                    curie_dict[prefix].add(identifier)
                
                if re.search('^rxn-|-rxn$', identifier, re.IGNORECASE):
                    prefix = 'metacyc.reaction'
                else:
                    prefix = 'metacyc.compound'
                
            elif re.fullmatch('^chebi$', extracted_curie[0], re.IGNORECASE):
                new_curie = extracted_curie[1].split(':')
                
                prefix = new_curie[0].upper()
                identifier = new_curie[1]
                
            else:
                if re.fullmatch('^brenda$', extracted_curie[0], re.IGNORECASE):
                    prefix = 'ec-code'
                else:
                    prefix = extracted_curie[0]
                
                identifier = extracted_curie[1]
                
        elif ':' in extracted_curie:
            extracted_curie = extracted_curie.split(':')
            
            if re.fullmatch('^biocyc$', extracted_curie[0], re.IGNORECASE) or ('metacyc.' in extracted_curie[0]):  # Check for bio- & metacyc
                prefix = 'biocyc'
                identifier = extracted_curie[-1]
                
                if not curie_dict or (prefix not in curie_dict):
                    curie_dict[prefix] = SortedSet()
                curie_dict[prefix].add(identifier)
                
                if re.search('^rxn-|-rxn$', identifier, re.IGNORECASE):
                    prefix = 'metacyc.reaction'
                else:
                    prefix = 'metacyc.compound'
                
            else:
                if re.fullmatch('^brenda$', extracted_curie[0], re.IGNORECASE):
                    prefix = 'ec-code'
                else:
                    prefix = extracted_curie[0]

                if re.fullmatch('^kegg.genes$', extracted_curie[0], re.IGNORECASE):
                    identifier = ':'.join(extracted_curie[1:len(extracted_curie)])
                else:
                    identifier = extracted_curie[1]
    
        # Use prefix as key & the corresponding set of identifiers as values   
        if not curie_dict or (prefix not in curie_dict):
            curie_dict[prefix] = SortedSet()

        curie_dict[prefix].add(identifier)
            
    return curie_dict

def generate_new_curie_set(prefix2id: SortedDict[str: SortedSet[str]], new_pattern: bool) -> SortedSet[str]: 
    """Generate a set of complete CURIEs from the provided prefix to identifier mapping
        
    Args:
        - prefix2id (SortedDict[str: SortedSet[str]]): Dictionary containing a mapping from database prefixes to their respective identifier sets 
        - new_pattern (bool):                          True if new pattern is wanted, otherwise False
            
    Returns:
        SortedSet: Sorted set containing complete CURIEs
    """
    curie_set = SortedSet()
    
    if new_pattern:
        SEPARATOR = ':'
    else:
        SEPARATOR = '/'
    
    for prefix in prefix2id:
        current_prefix = prefix   
        
        for identifier in prefix2id.get(prefix):
            separator = SEPARATOR

            if re.search('o$', prefix, re.IGNORECASE):  # Ontologies seem only to work with new pattern!
                separator = ':'
            
            elif re.fullmatch('^chebi$', current_prefix, re.IGNORECASE) and not new_pattern:  # The old pattern for chebi is different: Just adding '/' das NOT work!
                prefix = f'chebi/{current_prefix}'
                separator = ':'
                
            elif re.fullmatch('^biocyc$', prefix, re.IGNORECASE):  # Get identifier for biocyc
                prefix = f'biocyc{SEPARATOR}META'
                separator = ':'

            
            curie = MIRIAM + prefix + separator + identifier
            curie_set.add(curie)
            
    return curie_set

def add_curie_set(entity: SBase, qt, b_m_qt, curie_set: SortedSet[str]):
    """Add a complete CURIE set to the provided CVTerm
        
    Args:
        - entity (SBase):               A libSBML SBase object like model, GeneProduct, etc.
        - qt:                           A libSBML qualifier type: BIOLOGICAL_QUALIFIER|MODEL_QUALIFIER
        - b_m_qt:                       A libSBML biological or model qualifier type like BQB_IS|BQM_IS
        - curie_set (SortedSet[str]):   SortedSet containing CURIEs
    """        
    new_cvterm = generate_cvterm(qt, b_m_qt)
        
    for curie in curie_set:
        new_cvterm.addResource(curie)
            
    entity.addCVTerm(new_cvterm)

def improve_curie_per_entity(entity: SBase, new_pattern: bool):
    """Helper function: Removes duplicates & changes pattern according to new_pattern

    Args:
        - entity (SBase):      A libSBML SBase object, either a model or an entity
        - new_pattern (bool):  True if new pattern is wanted, otherwise False
    """
    not_miriam_compliant = []
    pattern = f'{MIRIAM}|{OLD_MIRIAM}'
    cvterms = entity.getCVTerms()
    
    for cvterm in cvterms:
        tmp_list = []
        
        # Retrieve QualifierType & Biological/ModelQualifierType before resource is removed!
        current_qt = cvterm.getQualifierType()
                
        if current_qt == BIOLOGICAL_QUALIFIER:
            current_b_m_qt = cvterm.getBiologicalQualifierType()
        elif current_qt == MODEL_QUALIFIER:
            current_b_m_qt = cvterm.getModelQualifierType()
            
        current_curies = [cvterm.getResourceURI(i) for i in range(cvterm.getNumResources())]
    
        for cc in current_curies:
            if re.match(pattern, cc, re.IGNORECASE):  # If model contains identifiers without MIRIAM/OLD_MIRIAM these are kept 
                tmp_list.append(cc)
                cvterm.removeResource(cc)
            else:
                not_miriam_compliant.append(cc)
            
        prefix2id = get_set_of_curies(tmp_list)
        curie_set = generate_new_curie_set(prefix2id, new_pattern)
        add_curie_set(entity, current_qt, current_b_m_qt, curie_set)
    
    if not_miriam_compliant:
        logging.info(f'The following {len(not_miriam_compliant)} entities are not MIRIAM compliant: {not_miriam_compliant}')

def improve_curies(entities: SBase, new_pattern: bool):
    """Removes duplicates & changes pattern according to new_pattern
    
    Args:
        - entities (SBase):     A libSBML SBase object, either a model or a list of entities
        - new_pattern (bool):  True if new pattern is wanted, otherwise False
    """
    if type(entities) == libModel:  # Model needs to be handled like entity!
        improve_curie_per_entity(entities, new_pattern)
    
    else: 
        for entity in tqdm(entities):
            improve_curie_per_entity(entity, new_pattern)
            
            if type(entity) == UnitDefinition:
                for unit in entity.getListOfUnits():  # Unit needs to be handled within ListOfUnitDefinition
                    improve_curie_per_entity(unit, new_pattern)


def polish_annotations(model: libModel, new_pattern: bool) -> libModel:
    """| Polishes all annotations in a model such that no duplicates are present 
       | & the same pattern is used for all CURIEs
        
    Args:
        - model (libModel):     Model loaded with libSBML
        - new_pattern (bool):   True if new pattern is wanted, otherwise False
        
    Returns:
        libModel: libSBML model with polished annotations
    """
    listOf_dict = {
        'model': model,
        'compartment': model.getListOfCompartments(),
        'metabolite': model.getListOfSpecies(),
        'parameter': model.getListOfParameters(),
        'reaction': model.getListOfReactions(),
        'unit definition': model.getListOfUnitDefinitions(),
        }

    if model.isPackageEnabled('fbc'):
        listOf_dict['gene product'] = model.getPlugin('fbc').getListOfGeneProducts()
    
    if model.isPackageEnabled('groups'):
        listOf_dict['group'] = model.getPlugin('groups').getListOfGroups()

    # Adjust annotations in model
    for listOf in listOf_dict:
        print(f'Polish {listOf} annotations...')
        improve_curies(listOf_dict[listOf], new_pattern)
    
    return model


def change_qualifier_per_entity(entity: SBase, new_qt, new_b_m_qt, specific_db_prefix: str=None) -> list:
    """Updates Qualifiers to be MIRIAM compliant for an entity

    Args:
        - entity (SBase): A libSBML SBase object like model, GeneProduct, etc.
        - new_qt (Qualifier): A libSBML qualifier type: BIOLOGICAL_QUALIFIER|MODEL_QUALIFIER
        - new_b_m_qt (QualifierType): A libSBML biological or model qualifier type like BQB_IS|BQM_IS
        - specific_db_prefix (str): Has to be set if only for a specific database the qualifier type should be changed. Can be 'kegg.genes', 'biocyc', etc.

    Returns:
        list: CURIEs that are not MIRIAM compliant
    """
    not_miriam_compliant = []
    pattern = f'{MIRIAM}|{OLD_MIRIAM}'
    cvterms = entity.getCVTerms()

    #for i in range(len(cvterms)):
    for cvterm in cvterms:
        tmp_set = SortedSet()
        #cvterm = cvterms.get(i)
        
        # include check for reaction and unit definition
        # if entity == Reaction or entity == UnitDefinition:
        # print(cvterm.getBiologicalQualifierType())
        if cvterm.getBiologicalQualifierType() == 9:  # 9 = BQB_OCCURS_IN (Reaction), Check for reactions with occursIn
            logging.info(f'CVTerm for {Fore.LIGHTYELLOW_EX}{str(entity)}{Style.RESET_ALL}' +
                  f' is left as {Fore.LIGHTYELLOW_EX}{BiolQualifierType_toString(cvterm.getBiologicalQualifierType())}{Style.RESET_ALL}')
        
        elif cvterm.getModelQualifierType() == 1:  # 1 = BQM_IS_DESCRIBED_BY (UnitDefinition), Check for UnitDefinitions with isDescribedBy
            logging.info(f'CVTerm for {Fore.LIGHTYELLOW_EX}{str(entity)}{Style.RESET_ALL}' + 
                  f' is left as {Fore.LIGHTYELLOW_EX}{ModelQualifierType_toString(cvterm.getModelQualifierType())}{Style.RESET_ALL}')
        
        else:
            current_curies = [cvterm.getResourceURI(j) for j in range(cvterm.getNumResources())]
        
            for cc in current_curies:
                    
                current_curie = None
                    
                if (specific_db_prefix != None) and (specific_db_prefix != ''):
                    if specific_db_prefix in cc:
                        current_curie = cc
                else:
                    current_curie = cc
                    
                if (current_curie) and re.match(pattern, current_curie, re.IGNORECASE):  # If model contains identifiers without MIRIAM/OLD_MIRIAM these are kept 
                    tmp_set.add(current_curie)
                    cvterm.removeResource(current_curie)
                else:
                    not_miriam_compliant.append(current_curie)
            
            add_curie_set(entity, new_qt, new_b_m_qt, tmp_set)
            #cvterms.remove(i)
                
    if not_miriam_compliant:
        return not_miriam_compliant


def change_qualifiers(model: libModel, entity_type: str, new_qt, new_b_m_qt, specific_db_prefix: str = None) -> libModel:
    """Updates Qualifiers to be MIRIAM compliant for an entity type of a given model 

    Args:
        - model (libModel):   Model loaded with libSBML
        - entity_type (str): Any string of the following: model|compartment|metabolite|parameter|reaction|unit definition|unit|gene product|group
        - new_qt (Qualifier): A libSBML qualifier type: BIOLOGICAL_QUALIFIER|MODEL_QUALIFIER
        - new_b_m_qt (QualifierType): A libSBML biological or model qualifier type like BQB_IS|BQM_IS
        - specific_db_prefix (str): Has to be set if only for a specific database the qualifier type should be changed. Can be 'kegg.genes', 'biocyc', etc.
    
    Returns:
        libModel: Model with changed qualifier for given entity type
    """
    not_miriam_compliant = []
    listOf_dict = {
        'model': model,
        'compartment': model.getListOfCompartments(),
        'metabolite': model.getListOfSpecies(),
        'parameter': model.getListOfParameters(),
        'reaction': model.getListOfReactions(),
        'unit definition': model.getListOfUnitDefinitions(),
        }
    
    if model.isPackageEnabled('fbc'):
        listOf_dict['gene product'] = model.getPlugin('fbc').getListOfGeneProducts()
    
    if model.isPackageEnabled('groups'):
        listOf_dict['group'] = model.getPlugin('groups').getListOfGroups()
        
    if entity_type == 'model':  # Model needs to be handled like entity!
        not_miriam_compliant = change_qualifier_per_entity(listOf_dict.get('model'), new_qt, new_b_m_qt, specific_db_prefix)
        
    elif entity_type == 'unit':
        for unit in listOf_dict.get('unit definition'):  # Unit needs to be handled within ListOfUnitDefinition
            not_miriam_compliant = change_qualifier_per_entity(unit, new_qt, new_b_m_qt, specific_db_prefix)
        
    else:
        try: 
            for entity in tqdm(listOf_dict.get(entity_type)):
                not_miriam_compliant = change_qualifier_per_entity(entity, new_qt, new_b_m_qt, specific_db_prefix)
        except(TypeError):
            logging.info('The entity ' +  entity_type + ' is not present in ' + model.getId())        
        
    if not_miriam_compliant:         
        print(f'The following {len(not_miriam_compliant)} entities are not MIRIAM compliant: {not_miriam_compliant}')
    
    return model


def change_all_qualifiers(model: libModel, lab_strain: bool) -> libModel:
    """Wrapper function to change qualifiers of all entities at once

    Args:
        - model (libModel): Model loaded with libSBML
        - lab_strain (bool): True if the strain was sequenced in a local lab

    Returns:
        libModel: Model with all qualifiers updated to be MIRIAM compliant
    """
    
    entity_list_mod = ['model',
                    'unit definition',
                    'unit']
    for entity in entity_list_mod:
        print(f'Change {str(entity)} qualifiers...')
        model = change_qualifiers(model, entity, MODEL_QUALIFIER, BQM_IS)
    
    entity_list = ['compartment',
                   'metabolite',
                   'parameter',
                   'reaction',
                   'gene product',
                   'group']
    for entity in entity_list:
        print(f'Change {str(entity)} qualifiers...')
        if lab_strain and entity == 'gene product':
            model = change_qualifiers(model, 'gene product', BIOLOGICAL_QUALIFIER, BQB_IS_HOMOLOG_TO)
        else:
            model = change_qualifiers(model, entity, BIOLOGICAL_QUALIFIER, BQB_IS)
        
    return model


#--------------------------------------------------- Main function ----------------------------------------------------#
def polish(model: libModel, email: str, id_db: str, protein_fasta: str, lab_strain: bool) -> libModel: 
    """| Completes all steps to polish a model
       | (Tested for models having either BiGG or VMH identifiers.)

    Args:
        - model (libModel): model loaded with libSBML
        - email (str): E-mail for Entrez
        - id_db (str): Main database identifiers in model come from
        - protein_fasta (str): File used as input for CarveMe
        - lab_strain (bool): True if the strain was sequenced in a local lab
    
    Returns:
        libModel: Polished libSBML model
    """
    colorama_init(autoreset=True)
    
    if lab_strain and not protein_fasta:
        print(Fore.LIGHTRED_EX + '''
                Setting the parameter lab_strain to True requires the provision of the protein FASTA file used as input for CarveMe.
                Otherwise, polish will not change anything for the GeneProducts.
                The header lines should look similar to the following line:
                >lcl|CP035291.1_prot_QCY37216.1_1 [gene=dnaA] [locus_tag=EQ029_00005] [protein=chromosomal replication initiator protein DnaA] [protein_id=QCY37216.1] [location=1..1356] [gbkey=CDS]
                It would also be a valid input if the header lines looked similar to the following line:
                >lcl|CP035291.1_prot_QCY37216.1_1 [locus_tag=EQ029_00005] [protein=chromosomal replication initiator protein DnaA] [protein_id=QCY37216.1]
                ''')
        return
    
    metab_list = model.getListOfSpecies()
    reac_list = model.getListOfReactions()
    gene_list = model.getPlugin('fbc').getListOfGeneProducts()

    ### unit definition ###
    add_fba_units(model)
    set_default_units(model)
    set_units(model)
    add_compartment_structure_specs(model)
    set_initial_amount(model)
    
    ## improve metabolite, reaction and gene annotations ###
    add_metab(metab_list, id_db)
    add_reac(reac_list, id_db)
    cv_notes_metab(metab_list)
    cv_notes_reac(reac_list)
    cv_ncbiprotein(gene_list, email, protein_fasta, lab_strain)
    
    ### set boundaries and constant ###
    polish_entities(metab_list, metabolite=True)
    polish_entities(reac_list, metabolite=False)

    
    ### MIRIAM compliance of CVTerms ###
    print('Remove duplicates & transform all CURIEs to the new identifiers.org pattern (: between db and ID):')
    polish_annotations(model, True)
    print('Changing all qualifiers to be MIRIAM compliant:')
    change_all_qualifiers(model, lab_strain)
    
    return model


#!/usr/bin/env python
""" Provides functions to investigate the model and test with MEMOTE

These functions enable simple testing of any model using MEMOTE and access to its number of reactions, metabolites and genes.
"""

import os
import memote
import json
import pandas as pd
import numpy as np
from cobra import Reaction
from libsbml import Model as libModel
from cobra import Model as cobraModel
from memote.support import consistency
# needed by memote.support.consitency
from memote.support import consistency_helpers as con_helpers
from refinegems.io import load_model_cobra, load_model_libsbml, search_sbo_label

__author__ = "Famke Baeuerle and Alina Renz"

DISSIPATION_RXNS = {
    'ATP':'atp_c + h2o_c --> adp_c + h_c + pi_c',
    'CTP':'ctp_c + h2o_c --> cdp_c + h_c + pi_c',
    'GTP':'gtp_c + h2o_c --> gdp_c + h_c + pi_c',
    'UTP':'utp_c + h2o_c --> udp_c + h_c + pi_c',
    'ITP':'itp_c + h2o_c --> idp_c + h_c + pi_c',
    'NADH':'nadh_c --> h_c + nad_c',
    'NADPH':'nadph_c --> h_c + nadp_c',
    'FADH2':'fadh2_c --> 2 h_c + fad_c',
    'FMNH2':'fmnh2_c --> 2 h_c + fmn_c',
    'Q8H2':'q8h2_c --> 2 h_c + q8_c',
    'MQL8':'mql8_c --> 2 h_c + mqn8_c',
    'DMMQL8':'2dmmql8_c --> 2 h_c + 2dmmq8_c',
    'ACCOA':'h2o_c + accoa_c --> h_c + ac_c + coa_c',
    'GLU':'glu__L_c + h2o_c --> akg_c + nh4_c + 2 h_c',
    'PROTON':'h_p --> h_c'
    }


def run_memote_sys(model: cobraModel):
    """Run MEMOTE on the local linux machine

    Args:
        - model (cobraModel): Model loaded with COBRApy
    """
    cmd = 'memote report snapshot ' + str(model)
    os.system(cmd)


def run_memote(model: cobraModel) -> dict:
    """Runs MEMOTE to obtain report as dict

    Args:
        - model (cobraModel): Model loaded with COBRApy

    Returns:
        dict: MEMOTE report as json in dict format
    """
    ret, res = memote.suite.api.test_model(model, sbml_version=None, results=True,
                                           pytest_args=None, exclusive=None, skip=None, 
                                           experimental=None, solver_timeout=10)
    snap = memote.suite.api.snapshot_report(res, html=False)
    result = json.loads(snap)
    return result

def get_memote_score(memote_report: dict) -> float:
    """Extracts MEMOTE score from report

    Args:
        - memote_report (dict): Output from run_memote.

    Returns:
        float: MEMOTE score
    """
    return memote_report['score']['total_score']


def initial_analysis(model: libModel) -> tuple[str, int, int, int]:
    """Extracts most important numbers of GEM

    Args:
        - model (libModel): Model loaded with libSBML

    Returns:
        tuple: Model name (1) & corresponding amounts of entities (2) - (4)
            (1) str: Name of model
            (2) int: Number of reactions 
            (3) int: Number of metabolites
            (4) int: Number of genes
    """
    name = model.getId()
    reactions = model.getNumReactions()
    metabolites = model.getNumSpecies()
    genes = len(model.getPlugin(0).getListOfGeneProducts())
    return name, reactions, metabolites, genes


def get_orphans_deadends_disconnected(model: cobraModel) -> tuple[list[str], list[str], list[str]]:
    """Uses MEMOTE functions to extract orphans, deadends and disconnected metabolites

    Args:
        - model (cobraModel): Model loaded with COBRApy

    Returns:
        tuple: Lists of metabolites that might cause errors (1) - (3) 
            (1) list: List of orphans
            (2) list: List of deadends
            (3) list: List of disconnected metabolites
    """
    orphans = consistency.find_orphans(model)
    deadends = consistency.find_deadends(model)
    disconnected = consistency.find_disconnected(model)

    orphan_list = []
    if len(orphans) > 0:
        for orphan in orphans:
            orphan_list.append(orphan.id)

    deadend_list = []
    if len(deadends) > 0:
        for deadend in deadends:
            deadend_list.append(deadend.id)

    disconnected_list = []
    if len(disconnected) > 0:
        for disc in disconnected:
            disconnected_list.append(disc.id)

    return orphan_list, deadend_list, disconnected_list


def get_mass_charge_unbalanced(model: cobraModel) -> tuple[list[str], list[str]]:
    """Creates lists of mass and charge unbalanced reactions,vwithout exchange reactions since they are unbalanced per definition

    Args:
        - model (cobraModel): Model loaded with COBRApy

    Returns:
        tuple: Lists of reactions that might cause errors (1) & (2)
        (1) list: List of mass unbalanced reactions
        (2) list: List of charge unbalanced reactions
    """

    mass_unbalanced = consistency.find_mass_unbalanced_reactions(
        model.reactions)
    charge_unbalanced = consistency.find_charge_unbalanced_reactions(
        model.reactions)

    mass_list = []
    if len(mass_unbalanced) > 0:
        for reac in mass_unbalanced:
            if (reac.id[:2] != 'EX'):
                mass_list.append(reac.id)

    charge_list = []
    if len(charge_unbalanced) > 0:
        for reac in charge_unbalanced:
            if (reac.id[:2] != 'EX'):
                charge_list.append(reac.id)

    return mass_list, charge_list


def get_model_info(modelpath: str) -> pd.DataFrame:
    """Reports core information of given model

    Args:
        - modelpath (str): Path to model file

    Returns:
        pd.DataFrame: Overview on model parameters
    """
    model_libsbml = load_model_libsbml(modelpath)
    model_cobra = load_model_cobra(modelpath)
    name, reac, metab, genes = initial_analysis(model_libsbml)
    orphans, deadends, disconnected = get_orphans_deadends_disconnected(
        model_cobra)
    mass_unbal, charge_unbal = get_mass_charge_unbalanced(model_cobra)
    model_info = pd.DataFrame([name,
                               reac,
                               metab,
                               genes,
                               orphans,
                               deadends,
                               disconnected,
                               mass_unbal,
                               charge_unbal],
                              ['model',
                               '#reactions',
                               '#metabolites',
                               '#genes',
                               'orphans',
                               'deadends',
                               'disconnected',
                               'mass unbalanced',
                               'charge unbalanced']).T

    return model_info


def parse_reaction(eq: str, model: cobraModel) -> dict:
    """Parses a reaction equation string to dictionary 

    Args:
        - eq (str): Equation of a reaction
        - model (cobraModel): Model loaded with COBRApy
        
    Returns:
       dict: Metabolite Ids as keys and their coefficients as values (negative = educts, positive = products)
    """
    # from Alina Renz
    eq = eq.split(' ')
    eq_matrix={}
    are_products = False
    coeff = 1
    for i,part in enumerate(eq):
        if part == '-->':
            are_products = True
            continue          
        if part == '+':
            continue
        if part == '2':
            coeff = 2
            continue
        if are_products:
            eq_matrix[model.metabolites.get_by_id(part)] = 1*coeff
            coeff = 1
        else:
            eq_matrix[model.metabolites.get_by_id(part)] = -1*coeff
            coeff = 1
    return eq_matrix


def get_egc(model: cobraModel) -> pd.DataFrame:
    """Energy-generating cycles represent thermodynamically infeasible states. Charging of energy metabolites without any energy source causes such cycles. Detection method is based on (Fritzemeier et al., 2017)

    Args:
        - model (cobraModel): Model loaded with COBRApy

    Returns:
        pd.DataFrame: Table with possible EGCs
    """
    dissipation_rxns = pd.DataFrame(DISSIPATION_RXNS.items(), columns=['type','equation'])
    
    with model: 
    # add dissipation reactions
        for i, row in dissipation_rxns.iterrows():
            try:
                met_atp = parse_reaction(row['equation'], model)
                rxn = Reaction(row['type'])
                rxn.name = 'Test ' + row['type'] + ' dissipation reaction'
                rxn.add_metabolites(met_atp)
                model.add_reaction(rxn)
            except(KeyError):
                dissipation_rxns.drop(dissipation_rxns[dissipation_rxns['type'] == row['type']].index, inplace=True)
            
        for rxn in model.reactions:
            if 'EX_' in rxn.id:
                rxn.bounds = (0.0, 0.0)
                #print('Set exchange rxn to 0', rxn.name)
            # set reversible reactions fluxes to [-1,1]    
            elif rxn.reversibility: 
                rxn.bounds = (-1.0,1.0)
                #print('Reversible rxn', rxn.name)
            # irreversible reactions have fluxes [0.1]    
            else:
                rxn.bounds = (0.0, 1.0)
                #print('Irreversible rxn', rxn.name)
                
        df_fluxes = pd.DataFrame()
        objval = dict()
        # optimize by choosing one of dissipation reactions as an objective
        for i, row in dissipation_rxns.iterrows():
            model.objective = row['type']
            #print('Set objective to', row['type'], ':', model.optimize().objective_value)
            objval[row['type']] = model.optimize().objective_value
            if model.optimize().objective_value > 0.0:
                df = pd.DataFrame.from_dict([model.optimize().fluxes]).T.replace(0, np.nan).dropna(axis=0).rename({'fluxes':row['type']}, axis=1)
                df_fluxes = pd.concat([df_fluxes, df], axis=1)
            else:
                df_fluxes[row['type']] = np.nan
        df_fluxes = pd.concat([df_fluxes,pd.DataFrame.from_dict([objval])])
    return df_fluxes.T.reset_index().rename({'index':'BOF', 0:'objective value'}, axis=1).fillna('')

def get_metabs_with_one_cvterm(model: libModel) -> list[str]:
    """Reports metabolites which have only one annotation, can be used as basis for further annotation research

    Args:
        - model (libModel): Model loaded with libSBML

    Returns:
        list: Metabolite Ids with only one annotation
    """
    spe = model.getListOfSpecies()

    only_one = [] #safe metab with only BiGG annotation
    for sb in spe:
        if sb.isSetId():
            pid = sb.getId()
            if sb.getCVTerm(0).getNumResources() == 1:
                only_one.append(pid)
                
    return only_one

def get_reactions_per_sbo(model: libModel) -> dict:
    """Counts number of reactions of all SBO Terms present

    Args:
        - model (libModel): Model loaded with libSBML

    Returns:
        dict: SBO Term as keys and number of reactions as values
    """
    sbos_dict = {}
    for react in model.getListOfReactions():
        sbo = react.getSBOTerm()
        if sbo in sbos_dict.keys():
            sbos_dict[sbo] += 1
        else: 
            sbos_dict[sbo] = 1
    return sbos_dict

def plot_rea_sbo_single(model: libModel):
    """Plots reactions per SBO Term in horizontal bar chart

    Args:
        - model (libModel): Model loaded with libSBML

    Returns:
        plot: Pandas Barchart
    """
    df = pd.DataFrame(get_reactions_per_sbo(model), index=[0]).T.reset_index().rename({0:model.id, 'index': 'SBO-Term'}, axis=1)
    df = df[df[model.id]>3]
    df['SBO-Name'] = df['SBO-Term'].apply(search_sbo_label)
    fig = df.drop('SBO-Term', axis=1).sort_values(model.id).set_index('SBO-Name').plot.barh(width=.8, figsize=(8,10))
    fig.set_ylabel('')
    fig.set_xlabel('number of reactions', fontsize=16)
    fig.legend(loc='lower right')
    return fig


#!/usr/bin/env python
""" Functions to enable annotation of entities using a manual curated table

While working on GEMs the user might come across ill-annotated or missing metabolites, reactions and genes. This module aims to enable faster manual curation by  allowing to edit an excel table directly which is used to update the given model. This module makes use of the cvterms module aswell.
"""
import pandas as pd
from tqdm.auto import tqdm
from libsbml import Model as libModel
from refinegems.cvterms import add_cv_term_reactions, add_cv_term_metabolites, metabol_db_dict, get_id_from_cv_term
from refinegems.entities import create_gpr_from_locus_tag, create_reaction

__author__ = "Famke Baeuerle"
    
            
def add_reactions_from_table(model: libModel, table: pd.DataFrame, email: str) -> libModel:
    """Wrapper function to use with table format given in data/manual_curation.xlsx, sheet gapfill: Adds all reactions with their info given in the table to the given model

    Args:
        - model (libModel): Model loaded with libSBML
        - table (pd-DataFrame): Table in format of sheet gapfill from manual_curation.xlsx located in the data folder
        - email (str): User Email to access the NCBI Entrez database

    Returns:
        libModel: Modified model with new reactions
    """
    for reaction_info in tqdm(table.groupby('BIGG')):
        reac_id = reaction_info[0]
        if model.getReaction(str(reac_id)) is None:
            reactants = dict(table.loc[table['BIGG'] == reac_id, ['educts', 'stoich_e']].dropna().values)
            products = dict(table.loc[table['BIGG'] == reac_id, ['products', 'stoich_p']].dropna().values)
            fluxes = table.loc[table['BIGG'] == reac_id, ['lower_bound', 'upper_bound']].dropna().to_dict('records')[0]
            name = table.loc[table['BIGG'] == reac_id, ['name']].dropna().to_dict('records')[0]['name']
            reversible = table.loc[table['BIGG'] == reac_id, ['reversible']].dropna().to_dict('records')[0]['reversible']
            fast = table.loc[table['BIGG'] == reac_id, ['fast']].dropna().to_dict('records')[0]['fast']
            try:
                sbo = table.loc[table['BIGG'] == reac_id, ['sbo']].dropna().to_dict('records')[0]['sbo']
            except (IndexError):
                print('SBO Term for ' + str(reac_id) + ' will be set to standard "SBO:0000167" (biochemical or transport reaction)')
                sbo = "SBO:0000167"
            reaction, model = create_reaction(model, reac_id, name, reactants, products, fluxes, reversible, fast, sbo)
            for (columnName, columnData) in table.loc[table['BIGG'] == reac_id].drop(['educts', 'stoich_e','products', 'stoich_p','lower_bound', 'upper_bound', 'name', 'reversible', 'fast'], axis=1).fillna(0).iteritems():
                for entry in columnData.values:
                    if not entry == 0:
                        if columnName == 'locus':
                            reaction.getPlugin(0).createGeneProductAssociation().createGeneProductRef().setGeneProduct(str(entry))
                            if model.getPlugin(0).getGeneProductByLabel(str(entry)) is None:
                                gpr, model = create_gpr_from_locus_tag(model, str(entry), email)
                        else:
                            add_cv_term_reactions(str(entry), str(columnName), reaction)
    return model


def update_annotations_from_table(model: libModel, table: pd.DataFrame) -> libModel:
    """Wrapper function to use with table format given in data/manual_curation.xlsx, sheet metabs: Updates annotation of metabolites given in the table

    Args:
        - model (libModel): Model loaded with libSBML
        - table (pd-DataFrame): Table in format of sheet metabs from manual_curation.xlsx located in the data folder

    Returns:
        libModel: Modified model with new annotations
    """
    table = table.drop(['Name', 'FORMULA', 'Notiz'], axis=1).fillna(0)
    table['PUBCHEM'] = table['PUBCHEM'].astype(int)
    for metab_info in tqdm(table.groupby('BIGG')):
        met = metab_info[0]
        for comp in ['_c', '_e', '_p']:
            try:
                metab = model.getSpecies('M_' + met + comp)
                #metab.unsetAnnotation()
                if not metab.isSetMetaId():
                    metab.setMetaId('meta_' + metab.getId())
                for (columnName, columnData) in table.loc[table['BIGG'] == met].iteritems():
                    for entry in columnData.values:
                        if not entry == 0:
                            add_cv_term_metabolites(str(entry), str(columnName), metab)
            except (AttributeError):
                print(met + comp + ' not in model')
    return model


def update_annotations_from_others(model: libModel) -> libModel:
    """Synchronizes metabolite annotations for core, periplasm and extracelullar

    Args:
        - model (libModel): Model loaded with libSBML

    Returns:
        libModel: Modified model with synchronized annotations
    """
    for metab in model.getListOfSpecies():
        base = metab.getId()[:-2]
        for comp in ['_c', '_e', '_p']:
            other_metab = model.getSpecies(base + comp)
            if other_metab is not None:
                if not other_metab.isSetMetaId():
                    other_metab.setMetaId('meta_' + other_metab.getId())
                for db_id, code in metabol_db_dict.items():
                    id = get_id_from_cv_term(metab, code)
                    for entry in id:
                        if entry is not None:
                            add_cv_term_metabolites(entry, db_id, other_metab)    
    return model               


# needed for installation of refinegems
from setuptools import setup

setup(name='refineGEMs',
      version='1.0.0',
      description='Scripts to curate a GEM',
      author='Famke Baeuerle',
      author_email='famke.baeuerle@gmail.com',
      license='MIT',
      packages=['refinegems'],
      install_requires = [
            "cobra==0.22.0",
            "biopython==1.79",
            "bioservices",
            "memote==0.13.0",
            "pandas==1.2.4",
            "numpy==1.20.3",
            # "pyyaml==5.4.1",
            "gffutils==0.10.1",
            "markupsafe==2.0.1", 
            "depinfo==1.7.0",
            "sortedcontainers==2.4.0",
            "libchebipy==1.0.10",
            "ratelimit==2.2.1",
            "sqlalchemy==1.4.43",
            "venn==0.1.3",
            "ols-client==0.1.3",
            "seaborn==0.12.2",
            "sqlalchemy==1.4.43",
            "click==8.1.3"
            ],
      zip_safe=False)



#!/usr/bin/env python
import re
import pandas as pd
from Bio import Entrez
from libsbml import Model as libModel
from libsbml import GeneProduct, Species, Reaction
from refinegems.cvterms import add_cv_term_genes, add_cv_term_metabolites, add_cv_term_reactions
from refinegems.io import search_ncbi_for_gpr
from typing import Union

__author__ = "Famke Baeuerle and Gwendolyn O. Gusak"


# Function originally from refineGEMs.genecomp/refineGEMs.KEGG_analysis --- Modified
def get_model_genes(model: libModel, kegg: bool=False) -> pd.DataFrame:
    """Extracts KEGG Genes/Locus tags from given model

    Args:
        - model (model-libsbml): Model loaded with libSBML
        - kegg (bool): True if KEGG Genes should be extracted, otherwise False

    Returns:
        pd.DataFrame: Table with all KEGG Genes/Locus tags in the model
    """
    genes_in_model = []
    for gene in model.getPlugin(0).getListOfGeneProducts():
        if kegg:
            cv_terms = gene.getCVTerms()
            if cv_terms:
                for cv_term in cv_terms:
                    for idx in range(cv_term.getNumResources()):
                        uri = cv_term.getResourceURI(idx)
                        if 'kegg.genes' in uri: genes_in_model.append(uri.split('kegg.genes:')[1])
        else:
            genes_in_model.append(gene.getLabel())

    return pd.DataFrame(genes_in_model)


# Function originally from refineGEMs.genecomp/refineGEMs.KEGG_analysis --- Modified
def compare_gene_lists(gps_in_model: pd.DataFrame, db_genes: pd.DataFrame, kegg: bool=True) -> pd.DataFrame:
    """Compares the provided tables according to column 0/'Locus_tag'
    
    Args:
        - gps_in_model (pd.DataFrame): Table containing the KEGG Gene IDs/Locus tags in the model
        - db_genes (pd.DataFrame): Table containing the KEGG Gene IDs for the organism from KEGG/
                                locus tags (Accession-2) from BioCyc
        - kegg (bool): True if KEGG Genes should be extracted, otherwise False
        
    Returns:
        pd.DataFrame: Table containing all missing genes
    """
    in_db = db_genes.set_index(0) if kegg else db_genes.set_index('locus_tag')
    in_model = gps_in_model.set_index(0)
        
    genes_in_db_not_in_model = in_db[~in_db.index.isin(in_model.index)]
    
    return genes_in_db_not_in_model.reset_index().iloc[:, 0] if kegg else genes_in_db_not_in_model.reset_index()


# Function originally from refineGEMs.genecomp/refineGEMs.KEGG_analysis --- Modified
def get_model_reacs_or_metabs(model_libsbml: libModel, metabolites: bool=False) -> pd.DataFrame:
    """Extracts table of reactions/metabolites with BiGG IDs from model

    Args:
        - model_libsbml (libModel): Model loaded with libSBML
        - metabolites (bool): Set to True if metabolites from model should be extracted

    Returns:
        pd.DataFrame: Table with BiGG IDs of reactions in the model
    """
    reac_or_metab_list = model_libsbml.getListOfSpecies() if metabolites else model_libsbml.getListOfReactions()

    list_of_reacs_or_metabs = []
    for reac_or_metab in reac_or_metab_list:
        list_of_reacs_or_metabs.append(reac_or_metab.id[2:])

    reac_or_metab_list_df = pd.Series(list_of_reacs_or_metabs)
    reac_or_metab_list_df = pd.DataFrame(reac_or_metab_list_df, columns=['bigg_id'])

    return reac_or_metab_list_df


def create_gpr_from_locus_tag(model: libModel, locus_tag: str, email: str) -> tuple[GeneProduct, libModel]:
    """Creates GeneProduct in the given model

    Args:
        - model (libModel): Model loaded with libSBML
        - locus_tag (str): NCBI compatible locus_tag
        - email (str): User Email to access the NCBI Entrez database

    Returns:
        tuple: libSBML GeneProduct (1) & libSBML model (2)
            (1) GeneProduct: Created gene product
            (2) libModel: Model containing the created gene product
    """
    Entrez.email = email
    name, locus = search_ncbi_for_gpr(locus_tag)
    gpr = model.getPlugin(0).createGeneProduct()
    gpr.setName(name)
    gpr.setId(locus_tag)
    gpr.setMetaId('meta_' + locus_tag)
    gpr.setLabel(locus_tag)
    gpr.setSBOTerm("SBO:0000243")
    add_cv_term_genes(locus_tag, 'NCBI', gpr)
    return gpr, model


def create_gp(model: libModel, model_id: str, name: str, locus_tag: str, protein_id: str) -> tuple[GeneProduct, libModel]:
    """Creates GeneProduct in the given model

    Args:
        - model (libModel): Model loaded with libSBML
        - model_id (str): ID identical to ID that CarveMe adds from the NCBI FASTA input file
        - name (str): Name of the GeneProduct
        - locus_tag (str): Genome-specific locus tag used as label in the model
        - protein_id (str): NCBI Protein/RefSeq ID

    Returns:
        tuple: libSBML GeneProduct (1) & libSBML model (2)
            (1) GeneProduct: Created gene product
            (2) libModel: Model containing the created gene product
    """
    id_db = None
    gp = model.getPlugin(0).createGeneProduct()
    gp.setId(model_id)
    gp.setName(name)
    gp.setLabel(locus_tag)
    gp.setSBOTerm('SBO:0000243')
    gp.setMetaId(f'meta_{model_id}')
    if re.fullmatch('^(((AC|AP|NC|NG|NM|NP|NR|NT|NW|WP|XM|XP|XR|YP|ZP)_\d+)|(NZ_[A-Z]{2,4}\d+))(\.\d+)?$', protein_id, re.IGNORECASE):
        id_db = 'REFSEQ'
    elif re.fullmatch('^(\w+\d+(\.\d+)?)|(NP_\d+)$', protein_id, re.IGNORECASE): id_db = 'NCBI'
    if id_db: add_cv_term_genes(protein_id, id_db, gp)
    return gp, model


def create_species(
    model: libModel, metabolite_id: str, name: str, compartment_id: str, charge: int, chem_formula: str
                   ) -> tuple[Species, libModel]:
    """Creates Species/Metabolite in the given model

    Args:
        - model (libModel): Model loaded with libSBML
        - metabolite_id (str): Metabolite ID within model (If model from CarveMe, preferable a BiGG ID)
        - name (str): Name of the metabolite
        - compartment_id (str): ID of the compartment where metabolite resides
        - charge (int): Charge for the metabolite
        - chem_formula (str): Chemical formula for the metabolite

    Returns:
        tuple: libSBML Species (1) & libSBML model (2)
        (1) Species: Created species/metabolite
        (2) libModel: Model containing the created metabolite
    """
    metabolite = model.createSpecies()
    metabolite.setId(f'M_{metabolite_id}')
    if name: metabolite.setName(name)
    metabolite.setMetaId(f'meta_M_{metabolite_id}')
    metabolite.setSBOTerm('SBO:0000247')
    metabolite.setInitialAmount(float('NaN'))
    metabolite.setHasOnlySubstanceUnits(True)
    metabolite.setBoundaryCondition(False)
    metabolite.setConstant(False)
    metabolite.setCompartment(compartment_id)
    metabolite.getPlugin(0).setCharge(charge)
    metabolite.getPlugin(0).setChemicalFormula(chem_formula)
    add_cv_term_metabolites(metabolite_id[:-2], 'BIGG', metabolite)
    return metabolite, model


def get_reversible(fluxes: dict[str: str]) -> bool:
    """Infer if reaction is reversible from flux bounds
    
    Args:
        - fluxes (dict): Dictionary containing the keys 'lower_bound' & 'upper_bound' 
                        with values in ['cobra_default_lb', 'cobra_0_bound', 'cobra_default_ub']
    Returns:
        bool: True if reversible else False
    """
    return (fluxes['lower_bound'] == 'cobra_default_lb') and (fluxes['upper_bound'] == 'cobra_default_ub')


def create_reaction(
    model: libModel, reaction_id: str, name:str, reactants: dict[str: int], products: dict[str: int], 
    fluxes: dict[str: str], reversible: bool=None, fast: bool=None, compartment: str=None, sbo: str=None, 
    genes: Union[str, list[str]]=None
    ) -> tuple[Reaction, libModel]:
    """Creates new reaction in the given model

    Args:
        - model (libModel): Model loaded with libSBML
        - reaction_id (str): BiGG ID of the reaction to create
        - name (str): Human readable name of the reaction
        - reactants (dict): Metabolites as keys and their stoichiometry as values
        - products (dict): Metabolites as keys and their stoichiometry as values
        - fluxes (dict): Dictionary with lower_bound and upper_bound as keys
        - reversible (bool): True/False for the reaction
        - fast (bool): True/False for the reaction
        - compartment (str): BiGG compartment ID of the reaction (if available)
        - sbo (str): SBO term of the reaction
        - genes (str|list): List of genes belonging to reaction

    Returns:
        tuple: libSBML reaction (1) & libSBML model (2)
            (1) Reaction: Created reaction 
            (2) libModel: Model containing the created reaction
    """
    reaction = model.createReaction()
    reaction.setId('R_' + reaction_id)
    if name: reaction.setName(name)
    reaction.setMetaId('meta_R_' + reaction_id)
    sbo = sbo if sbo else 'SBO:0000167'  # SBO term for biochemical or transport reaction
    reaction.setSBOTerm(sbo)
    fast = fast if fast else False
    reaction.setFast(fast)
    if compartment: reaction.setCompartment(compartment)  # Set compartment for reaction if available
    reversible = reversible if reversible else get_reversible(fluxes)
    reaction.setReversible(reversible)
    if genes:
            if genes == 'G_spontaneous':
                reaction.getPlugin(0).createGeneProductAssociation().createGeneProductRef().setGeneProduct(gene)
            elif len(genes) == 1:
                reaction.getPlugin(0).createGeneProductAssociation().createGeneProductRef().setGeneProduct(genes[0])
            else:
                gp_ass_or = reaction.getPlugin(0).createGeneProductAssociation().createOr()
                for gene in genes:
                    # Set GeneProductReferences if available
                    gp_ass_or.createGeneProductRef().setGeneProduct(gene)
    for metab, stoich in reactants.items(): #reactants as dict with metab:stoich
        reaction.addReactant(model.getSpecies('M_' + metab), stoich)
    for metab, stoich in products.items(): #reactants as dict with metab:stoich
        reaction.addProduct(model.getSpecies('M_' + metab), stoich)
    reaction.getPlugin(0).setLowerFluxBound(fluxes['lower_bound'])
    reaction.getPlugin(0).setUpperFluxBound(fluxes['upper_bound'])
    add_cv_term_reactions(reaction_id, 'BIGG', reaction)
    return reaction, model
 
